#include <assert.h>
#include <iostream>
#include <algorithm>
#include "aesanode.h"
#ifdef MYERS_DIST
    #include "myers.hpp"
#else
    #include "edist.h"
#endif
#include "myagrep.h"

typedef vector<StringList> VStringList;


#ifdef MYERS_DIST
CMyersEdistFastPair EdCalc;
#endif


// Do not change this
const char CAESANodeBase::termChar = 0;

CAESANodeBase::~CAESANodeBase() {} // An odd pure virtual C++ destructor

void CAESANode::Insert(const StringList & lst, bool hasTransp, int BucketSize, bool silent, int level)
{
    if (!silent) {
        std:: cout << " Insert of " << lst.size () << " strings, level = "<<level<<" this = "<<this << std::endl;
    }

    // First of all we choose the root randomly if not exists
    unsigned int ListSize = lst.size ();
    if (!ListSize)
        return;
    if (!mvpNodes.size ()) {
        int num = 0;
        
        num = rand () % ListSize;

        mRoot = lst[num];

    }

    VStringList SplitList;
    StringList  EmptyList;

    for (StringList::const_iterator it = lst.begin (); it != lst.end (); it++) {
#ifdef MYERS_DIST
        int dst = EdCalc.CalcDist(mRoot, *it, hasTransp);
#else
        int dst = EditDistance(mRoot, *it, hasTransp);
#endif
        if (dst > 0) {
            if (dst > mvpNodes.size()) {
                mvpNodes.resize(dst);
                SplitList.resize(dst);
            }

            SplitList[dst - 1].push_back(*it);
        }
    }

    // Now the lists are split and must be inserted into the nodes
    // 1. If the size of the list is smaller than BucketSize, create a node of type CAESABucket
    // 2. Otherwise, create a regular node (it will create AESA tree recursively)
    for (int i = 0; i < mvpNodes.size() ; i++) {
        if (SplitList[i].size()) {
            mvpNodes[i] = SplitList[i].size() <= BucketSize ? static_cast<CAESANodeBase*>(new CAESABucket) : static_cast<CAESANodeBase*>(new CAESANode);
            mvpNodes[i]->Insert(SplitList[i], hasTransp, BucketSize, silent, level + 1);
        }
    }

}

size_t CAESANode::DumpToFile(off_t offset, int fd)
{
    // We writing data in the following order
    // 1. K - number of nodes u_int8_t 
    // 2. root (which might be a sequence of words)
    // 3. K times subtree records in the form
    //   subtree distance from the root (u_int8_t)
    //   offset of the subtree
    // 4. K - subtrees
    // 5. Update pointers to subtrees
    u_int8_t K = 0;

    size_t buflen = sizeof (K) +  mRoot.size() +  1;

    for (size_t i = 0; i < mvpNodes.size (); i++) {
        if (mvpNodes[i]) {
            K++;
            buflen += sizeof (offset) + sizeof (u_int8_t);
        }
    }

    int delta = 0;
    if (offset) { delta = (offset % getpagesize ()); }

    if (lseek (fd, offset + buflen - 1, SEEK_SET) < 0 || write (fd, "", 1) < 1) {
        return INVALID_SIZE;
    }

    void *buf = mmap (NULL, buflen + delta, PROT_WRITE | PROT_READ, MAP_SHARED, fd, offset - delta);

    if (!buf || buf == MAP_FAILED) {
        return INVALID_SIZE;
    }

    char *sbuf = (char *) buf + delta;
    // Step 1.
    *(u_int8_t *) (void *) sbuf = K;
    sbuf += sizeof (K);
    // Step 2.
    memcpy (sbuf, mRoot.c_str(), mRoot.size() + 1);
    sbuf += mRoot.size() + 1;
    
    // Step 3.
    off_t *ptrs[K];
    int cnt = 0;

    for (u_int8_t i = 0; i < mvpNodes.size (); i++) {
        if (mvpNodes[i]) {
            *(u_int8_t *) sbuf = i;
            sbuf += sizeof(u_int8_t);
            ptrs[cnt++] = (off_t *) sbuf;
            assert (cnt <= K);
            sbuf += sizeof (off_t);
        }
    }

    int sub_size = 0;
    // Step 4.
    cnt = 0;
    for (size_t i = 0; i < mvpNodes.size (); i++)
    if (mvpNodes[i]) {
        off_t ChildOffset = offset + buflen + sub_size;
        size_t s_sz = mvpNodes[i]->DumpToFile (ChildOffset, fd);
        if (s_sz == INVALID_SIZE) {
            munmap (buf, buflen + delta);
            return INVALID_SIZE;
        }
        *ptrs[cnt++] = ChildOffset; 
        assert (cnt <= K);
        sub_size += s_sz;
    }


    assert (sbuf - (char *) buf == delta + buflen);

    if (munmap (buf, buflen + delta) < 0) {
        return INVALID_SIZE;
    }

    return buflen + sub_size;
}

void CAESANode::SearchAtLevel (const std::string & s, int nerr, 
                             const myagrep & a, 
#ifdef MYERS_DIST
                             CMyersEdist& MyersEdist, 
#endif
                             int SeqCutoff, 
                             int *numComp, StringList * resList, bool hasTransp,
                             bool silent)
{
    if (numComp)
        (*numComp)++;

    // We must check the root
#ifdef MYERS_DIST
    int d = EdCalc.CalcDist (s, mRoot, hasTransp);
#else
    int d = EditDistance(s, mRoot, hasTransp);
#endif

    if (d <= nerr) {
        if (!silent)
            cout << mRoot << "\n";
        if (resList != NULL)
            resList->push_back(mRoot);
    }
    int imin = d - nerr, imax = d + nerr;
    if (imin < 1)
        imin = 1;
    if (imax > mvpNodes.size())
        imax = mvpNodes.size();

    --imin;
    --imax;

    for (int i = imin; i <= imax; i++)
    if (mvpNodes[i]) {
        mvpNodes[i]->SearchAtLevel (s, nerr, a, 
        #ifdef MYERS_DIST
                                    MyersEdist, 
        #endif
                                    SeqCutoff, numComp, resList, hasTransp, silent);
    }
}

void CAESANode::Print (int level)
{
    for (int i = 0; i < level; i++)
        cout << "*";
    cout << ":" << mRoot << "\n";

    for (int i = 0; i < mvpNodes.size(); i++) {
        if (mvpNodes[i]) {
            for (int j = 0; j < level; j++)
                cout << "  ";
            cout << " dst " << (i + 1) << "\n";
            mvpNodes[i]->Print (level + 1);
        }
    }
}


CAESANode::~CAESANode () {
    for (int i = 0; i < mvpNodes.size(); i++)
        if (mvpNodes[i]) {
            delete mvpNodes[i];
        }
}


void CAESABucket::Print (int level)
{
    for (int i = 0; i < level; i++)
        cout << "*";

    for (int i = 0; i < mvList.size(); i++) {
        cout << " " << mvList[i];
    }
}

void CAESABucket::Insert(const StringList & lst, bool hasTransp, int BucketSize, bool silent, int level)
{
    if (!silent) {
        std:: cout << " Insert of " << lst.size () << " strings, level = "<<level<<" this = "<<this << std::endl;
    }

    size_t N = lst.size();
    mvList.assign(lst.begin(), lst.end());
    mpDistMatr = new u_int8_t[GetMatrSize(N)];


    for (int i = 0; i < N; ++i) {
        for (int j = 0; j < i; ++j) {
#ifdef MYERS_DIST
            int d = EdCalc.CalcDist (mvList[i], mvList[j], hasTransp);
#else
            int d = EditDistance(mvList[i], mvList[j], hasTransp);
#endif
            GetDistCell(i, j) = d;
        }
        GetDistCell(i, i) = 0;
    }
}

size_t CAESABucket::DumpToFile(off_t offset, int fd)
{
    if (lseek (fd, offset, SEEK_SET) < 0) {
        return INVALID_SIZE;
    }

    // 1. Write string qty
    size_t buflen = 0;
    size_t N = mvList.size();

    if (write(fd, &N, sizeof N) != sizeof N) {
        return INVALID_SIZE;
    }
    buflen += sizeof N;

    // 2. Write strings themselves 
    for (size_t i = 0; i < mvList.size(); ++i) {
        if (write(fd, mvList[i].c_str(), mvList[i].size() + 1) != mvList[i].size() + 1) {
            return INVALID_SIZE;
        }
        buflen += mvList[i].size() + 1;
    }

    // 3. Write a distance array
    size_t sz = GetMatrSize(N) * sizeof *mpDistMatr;

    if (write(fd, mpDistMatr, sz) != sz) {
        return INVALID_SIZE;
    }

    buflen += sz;

    return buflen;
}


void CAESABucket::SearchAtLevel (const std::string & s, int nerr, 
                             const myagrep & a, 
#ifdef MYERS_DIST
                             CMyersEdist& MyersEdist, 
#endif
                             int SeqCutoff,
                             int *numComp, StringList * resList, bool hasTransp,
                             bool silent)
{
    if (numComp)
        (*numComp)++;

    size_t N = mvList.size();

    if (!N) {
        return;
    }

    size_t      IdStr[2][N];
    int         NLeft = N;

    size_t i = 0; 

    assert(N);
    // making IdStr[i] == i
#undef STEP
#define STEP assert(i<N);IdStr[0][i] = i++;
#define UNFOLD_LOOP 1

#if UNFOLD_LOOP
    switch ( N & 7) {
        case 0: while (i < N) {
                    STEP
        case 7:     STEP
        case 6:     STEP
        case 5:     STEP
        case 4:     STEP
        case 3:     STEP
        case 2:     STEP
        case 1:     STEP
                }
    }
#else 
    while (i<N) {
        STEP
    }
#endif


    if (SeqCutoff <= 0) SeqCutoff = 1; 

    int         nCurr = 0;
    while (NLeft > SeqCutoff) {
        // Always compare the string with the last element in the list
        size_t PivotId = rand() % NLeft; /*IdStr[nCurr][NLeft - 1];*/
        const string& pivot = mvList[PivotId];
#ifdef MYERS_DIST
        int d = MyersEdist.CalcDist (pivot.c_str());
#else
        int d = EditDistance(s, pivot, hasTransp);
#endif
        if (d <= nerr) {
            if (!silent)
                cout << pivot << "\n";
            if (resList != NULL)
                resList->push_back(pivot);
        }
        int nNext = 1 - nCurr;
        int nNextLeft = 0;

        int dMin = d - nerr;
        int dMax = d + nerr;

        --NLeft;

        int i = NLeft - 1;

        if (NLeft > 1) {
            size_t CurrId;
            int    d1;
            #undef STEP
            #define STEP assert(i>=0);CurrId = IdStr[nCurr][i--]; d1= GetDistCell(CurrId, PivotId); if (d1 >= dMin && d1 <= dMax) { IdStr[nNext][nNextLeft++] = CurrId; }

#if UNFOLD_LOOP
            switch ((i+1) & 7) {
                case 0: while (i > 0) {
                            STEP
                case 7:     STEP
                case 6:     STEP
                case 5:     STEP
                case 4:     STEP
                case 3:     STEP
                case 2:     STEP
                case 1:     STEP
                        }
            }
#else
        while (i >= 0) { STEP; }
#endif
        }
        NLeft = nNextLeft;
        nCurr = nNext;
    }

    if (NLeft) {
        if (numComp)
            (*numComp) += NLeft;

        size_t TotalStrSize = 0;

        int i = NLeft - 1;
        size_t*     pIds = IdStr[nCurr];

#undef STEP
#define STEP assert(i>=0);TotalStrSize += mvList[pIds[i--]].size() + 1;

#if UNFOLD_LOOP
        switch ( (i+1) & 7) {
            case 0:  while (i > 0) {
                        STEP
            case 7:     STEP
            case 6:     STEP
            case 5:     STEP
            case 4:     STEP
            case 3:     STEP
            case 2:     STEP
            case 1:     STEP
                    }
        }
#else
         while (i >= 0) {
            TotalStrSize += mvList[pIds[i--]].size() + 1;
         }
#endif

        char pBuff[TotalStrSize];

        if (!NLeft) {
            return;
        }

        i = NLeft - 1;
        char *p = pBuff;

        #undef  STEP
        #define STEP  { const string& s = mvList[pIds[i--]]; size_t sz = s.size() + 1 ; memcpy(p, s.c_str(), sz); p += sz; }

#if UNFOLD_LOOP
        switch ( (i+1) & 7) {
                case 0: while (i > 0) {
                            STEP
                case 7:     STEP
                case 6:     STEP
                case 5:     STEP
                case 4:     STEP
                case 3:     STEP
                case 2:     STEP
                case 1:     STEP
                        }
        }
#else
        while (i >= 0) { 
            STEP
        }
#endif

#ifdef MYERS_DIST
        if (nerr > 2) {
                const char* pBeg    = pBuff;
                const char* pEnd    = pBuff + TotalStrSize;
                const char* pCurr   = pBeg;

                for (const char *p = pBeg; pBeg < pEnd; pBeg = p + 1) {
                    p = strchr(pBeg, termChar);

                    if (!p) {
                        cerr << "Internal error: expect TermChar" << endl;
                        abort();
                    }


                    int dist = MyersEdist.CalcDist(pBeg);
        
                    if (dist <= nerr) {
                        if (!silent)
                            cout << pBeg << "\n";
                        if (resList != NULL)
                            resList->push_back(pBeg);
                    }
            }
            return;
        }
#endif
        a.findPattern(CInterface::filterType, termChar, pBuff, pBuff + TotalStrSize, silent, resList);
        return;
    }

}

CAESABucket::~CAESABucket() 
{
    delete [] mpDistMatr; 
}
