#include <iostream>
#include <string>
#include <vector>
#include <fstream>

#include "hashport.h"
#include "trieex.h"
#include "alphabet.h"
#include "utils.h"

#include "interface.h"
#include "../StringTriesNew/trie.h"

using namespace std;


typedef vector<string> vs;

DECLARE_STATIC_COUNTER(TestCnt)

template <typename TIndex>
void
TestIndex(TIndex& indx, const vs& vPatList) {
    RESET_COUNTER(TestCnt)

    unsigned sz = vPatList.size();
    unsigned FoundQty = 0;

    START_COUNTER(TestCnt)
    for (unsigned i = 0; i < sz; ++i) {
        FoundQty += indx.Check(vPatList[i]);
    }
    STOP_COUNTER(TestCnt)

    cout << "Index type: " << indx.Name() << endl;
    cout << "    Searched for " << sz << " patterns, found: " << FoundQty << endl;
    cout << "    Total time: " << 
             GET_COUNTER(TestCnt) << " Avg lookup time: " << double(GET_COUNTER(TestCnt)) / sz << endl;
}

class CHashWrap {
public:
    void Insert(const string& s) { mData.insert(s); }
    bool Check(const string& s) { return mData.find(s) != mData.end(); }
    string Name() {  return "Hash"; }
private:
    typedef hash_set<string>    hs;

    hs                          mData;
};

class CTrieWrap {
public:
    void Insert(const string& s) { mTrie.Insert(s + EOS); }
    bool Check(const string& s) { 

#if 0
        StringList  lst;
        mTrie.SearchExact(s, NULL, &lst, true);
        return lst.size() == 1 && lst[0] == s;
#else
        // Testing without any memory allocations
        return mTrie.SearchExact(s, NULL, NULL, true);
#endif
    }
    string Name() {  return "Path-Compressed Trie"; }
private:
    CTrie   mTrie;
};

void Usage(const char* pProg)
{
    cerr << "Usage: " << pProg << " <dict file> <pattern file> " << endl;
}

int
main(int argc, char* pArgv[])
{
    string              s;

    GetEnv(ENV_ABC, s);

    EAlphabet           AlphaType = GetAlphabetType(s);

    if (AlphaType == eUnknown) {
        cerr << "Unknown alphabet: '" << s << "' check env var " << ENV_ABC << endl;
        return 1;
    }

    if (argc != 3) {
        Usage(pArgv[0]);
        return 1;
    }

    CTrieExact  IndexTrie(GetAlphabet(AlphaType));
    CHashWrap   IndexHash;
    CTrieWrap   IndexTrieComp;

    const char* pDictFile = pArgv[1];
    const char* pPatFile = pArgv[2];

    ifstream    fDict(pDictFile);
    ifstream    fPat(pPatFile);

    if (!fDict) {
        cerr << "Cannot open '" << pDictFile << "'" << endl;
        return 1;
    }
    if (!fPat) {
        cerr << "Cannot open '" << pPatFile << "'" << endl;
        return 1;
    }

    try {

    unsigned DictQty = 0;

    for (;getline(fDict, s); ++DictQty) {
        IndexTrie.Insert(s);
        IndexTrieComp.Insert(s);
        IndexHash.Insert(s);
    }

    if (fDict.bad()) {
        cerr << "Error reading from '" << pDictFile << "'" << endl;
        return 1;
    }

    vs  vPatList;

    while (getline(fPat, s)) {
        vPatList.push_back(s);
    }

    if (fPat.bad()) {
        cerr << "Error reading from '" << pPatFile << "'" << endl;
        return 1;
    }

    cout << "Inserted " << DictQty << " strings, read " << vPatList.size() << " patterns." << endl;

    TestIndex(IndexTrie, vPatList);

    cout << "Trie consumed: " << CTrieExact::GetUsedMem() / 1024 << " Kb(s)" << endl;
    TestIndex(IndexTrieComp, vPatList);
    TestIndex(IndexHash, vPatList);

    } catch (const string& err) {
        cerr << "Error: " << err << endl;
        return 1;
    }

    return 0;
}
