/*
 *  A memory efficient trie. 
 *  The trie is designed in a way that decreases memory fragmentation.
 *  It uses a separate memory allocator. It may take advantage of the
 *  fact that trie uses small chunks of memory.
 *  1) Whenever a trie node changes its size,
 *  a new memory block is allocated. The old memory block
 *  is returned to the memory allocator.
 *  2) To avoid updating parent nodes, nodes are addresssed
 *  by a 4-byte index, rather than a pointer. A special 
 *  table that maps indices into pointers is maintained.
 *
 */
#ifndef COMPTRIE_H
#define COMPTRIE_H

#include <assert.h>
#include <string.h>

#include <iostream>
#include <sstream>
#include <vector>
#include <string>
#include <stack>
#include <stdexcept>
#include <algorithm>
#include <../Common/myexcept.h>
#include "nodehelper.h"

#define TEST_MODE
#ifdef TEST_MODE
#   define WILDCHARD_CHAR '?'
#else
#   define WILDCHARD_CHAR '\01'
#endif

inline bool WildCardEqual(char c1, char c2) {
    // We don't match the wildcard against the end of the string
    return c1 == c2 || (c1 == WILDCHARD_CHAR && c2) || (c2 == WILDCHARD_CHAR && c1);
}

struct CEntry {
    NodeIndex      mCurrIndex;
    SmallCounter   mCurrChild;
    unsigned int   mLevel;

    CEntry(NodeIndex CurrIndex = 0, SmallCounter CurrChild = 0, unsigned int level = 0) : 
            mCurrIndex(CurrIndex), mCurrChild(CurrChild), mLevel(level) {}
};

/*
 *  Trie has the following properties, see CheckerVisitor in testutil.h
 *
 *  1) The root of this trie is always a node with an empty label.  Otherwise, Insert and Merge operations may not work correctly.
 *  2) Some terminal nodes may have empty labels. This happens, when one string is a proper prefix of another string.
 *  3) If a node is neither a root node nor a terminal node, its labels is not empty.
 *
 */
template <typename NodeAllocator>
class CCompTrie {
public:
    CCompTrie(NodeIndex RootIndex = NILL_NODE_INDEX) : mRootIndex(RootIndex) {}
    void Insert(const std::string& Str);

    template <typename Visitor> void VisitTrie(Visitor& ) const;
    inline void GetNodeInfo(NodeIndex Index, SmallCounter &ChildQty, SmallCounter &LabelLen, const char*& pLabel, const NodeIndex*& pChildLabelLen) const;
    bool Empty() const { return mRootIndex == NILL_NODE_INDEX; }
    NodeIndex GetRoot() const { 
                                if (Empty()) { 
                                    THROW_AND_PRINT_RUNTIME_ERROR("Trie is empty!"); 
                                }
                                return mRootIndex; 
                              }

    static NodeIndex Clone(NodeIndex StartNode, bool bReuseExistingNodes = false);     
    static NodeIndex Clone(const CCompTrie& trie, bool bReuseExistingNodes = false) { 
        return !trie.Empty() ? Clone(trie.GetRoot(), bReuseExistingNodes) : NILL_NODE_INDEX; 
    }

    // . Returning by value is efficient here, because the only class member is an integer
    CCompTrie CloneTrie() const { return CCompTrie(Clone(*this)); }

    // 1. Merges tries, 
    // 2. If bReuseExistingNodes == true, then a procedure reuses existing Nodes. 
    //    Therefore, memory of merged tries cannot be freed. Also note that original tries remain unchanged. 
    //    Otherwise, the procedure will create brand new nodes.
    // 3. Returning by value is efficient here, because the only class member is an integer
    CCompTrie     MergeTries(const CCompTrie &OthTree, bool bReuseExistingNodes) const;
    /*CCompTrie     MergeTries(NodeIndex OthTreeRoot, bool bReuseExistingNodes) const {
        return CCompTrie<NodeAllocator>(OthTreeRoot);
    }*/

    void AccessNode(NodeIndex Index, SmallCounter& LabelLen, const char*& pLabel, SmallCounter& ChildQty, NodeIndex*& pChildren);
    void AccessNode(NodeIndex Index, SmallCounter& LabelLen, const char*& pLabel, SmallCounter& ChildQty, const NodeIndex*& pChildren) const;
    void ConcatLabels(const NodeIndex* pVisitedNodes, SmallCounter Qty, std::string& res) const;

    static NodeIndex     CreateInsertionTrie(NodeIndex Index, SmallCounter Centroid);
    static NodeIndex     CreateSubstitutionTrie(NodeIndex Index, SmallCounter Centroid);
    static NodeIndex     CreateDeletionTrie(NodeIndex Index, SmallCounter Centroid);
    static NodeIndex     CreateTranspositionTrie(NodeIndex Index, SmallCounter Centroid);
    static NodeIndex     CreateTrieWithOneChild(NodeIndex OneChild);
    static NodeIndex     CreateTrieWithNoChild();
    static bool IsLongString(const std::string& s) { return s.size() >= MAX_SMALL_COUNTER; }

    void   AddToArrayIfNotEmpty(std::vector<NodeIndex>& vArray) {
        if (!Empty()) { 
            if (!vArray.empty() && vArray[vArray.size() - 1] == GetRoot()) { return; }
            vArray.push_back(GetRoot()); 
        }
    }
    // Creates a trie, attached to the given label.
    CCompTrie AppendTrieToEdge(const std::string& Label, bool bReuseExistingNodes) const;
private:
    SmallCounter  CheckCommon(const char *pStr1, SmallCounter Len1, const char *pStr2, SmallCounter Len2) const;
    SmallCounter  CheckCommon(NodeIndex Index, const std::string &Str, SmallCounter Start) const;

    NodeIndex     MergeTriesInternal(NodeIndex RootIndex1, NodeIndex RootIndex2, bool bReuseExistingNodes) const;
    NodeIndex     mRootIndex;
};

template <typename NodeAllocator>
void 
CCompTrie<NodeAllocator>::ConcatLabels(const NodeIndex* pVisitedNodes, SmallCounter Qty, std::string& res) const
{
    res.clear();
    for (SmallCounter i = 0; i < Qty; ++i) {
        NodeIndex      Index = pVisitedNodes[i];

        if (Index >= NodeAllocator::mpNodePtr.size()) {
            THROW_AND_PRINT_RUNTIME_ERROR("Invalid node index " << Index);
        }
        CNode*          pNode = NodeAllocator::mpNodePtr[Index];
        res += std::string(pNode->mpLabel, pNode->mLabelLen);
    }
}

template <typename NodeAllocator>
CCompTrie<NodeAllocator>
CCompTrie<NodeAllocator>::AppendTrieToEdge(const std::string& Label, bool bReuseExistingNodes) const
{
    if (Empty()) {
        THROW_AND_PRINT_RUNTIME_ERROR("Does not work with empty tries");
    }
    if (IsLongString(Label)) {
        THROW_AND_PRINT_RUNTIME_ERROR("Label '" << Label << "' is too long");
    }

    CNode*          pRootNode = NodeAllocator::mpNodePtr[mRootIndex];
    if (pRootNode->mLabelLen) {
        THROW_AND_PRINT_RUNTIME_ERROR("Invalid root index: " << mRootIndex << " label is non-empty");
    }

    NodeIndex   NewIndex = NILL_NODE_INDEX;
    NodeIndex*  pRootChildren = pRootNode->GetChildData();

    if (Label.empty()) {
        return NodeAllocator::AllocNode(0, "", pRootNode->mChildQty, pRootChildren);
    } else {
        if (pRootNode->mChildQty > 1) {
            NewIndex = NodeAllocator::AllocNode(static_cast<SmallCounter>(Label.size()), Label.c_str(), pRootNode->mChildQty);
        } else {
            std::string  Label1 = Label;

            if (pRootNode->mChildQty) {
                assert(pRootNode->mChildQty == 1);
                CNode*  pChildNode = NodeAllocator::mpNodePtr[pRootChildren[0]];

                Label1 += std::string(pChildNode->mpLabel, pChildNode->mLabelLen);
            }

            NewIndex = NodeAllocator::AllocNode(static_cast<SmallCounter>(Label1.size()), Label1.c_str(), pRootNode->mChildQty);
        }
    }

    CNode*          pNewNode = NodeAllocator::mpNodePtr[NewIndex];
    NodeIndex*      pNewChildren = pNewNode->GetChildData();

    for (SmallCounter i = 0; i < pRootNode->mChildQty; ++i) {
        pNewChildren[i] = Clone(pRootChildren[i], bReuseExistingNodes);
    }

    return CreateTrieWithOneChild(NewIndex);
}

template <typename NodeAllocator>
NodeIndex
CCompTrie<NodeAllocator>::CreateTranspositionTrie(NodeIndex StartNode, SmallCounter Centroid)
{
    if (StartNode >= NodeAllocator::mpNodePtr.size()) {
        THROW_AND_PRINT_RUNTIME_ERROR("Invalid node index " << StartNode);
    }

    CNode*          pOldNode = NodeAllocator::mpNodePtr[StartNode];
    NodeIndex*      pChildren = pOldNode->GetChildData();

    SmallCounter    NewChildCnt = 0;

    CCompTrie<NodeAllocator> FirstTrie ; // It MUST BE AN EMPTY trie, i.e. no children at all

    for (SmallCounter ChildId = 0; ChildId < pOldNode->mChildQty; ++ChildId) 
    if (ChildId != Centroid) {
        CNode*      pChild = NodeAllocator::mpNodePtr[pChildren[ChildId]];

        SmallCounter    ChildLabelLen = pChild->mLabelLen; 
        NodeIndex*      pGrandChildren = pChild->GetChildData();

        if (ChildLabelLen > 1) {
        // If label is long enough, just transpose label's first two characters

            NodeIndex   SubTreeIndex = NodeAllocator::AllocNode(ChildLabelLen, pChild->mpLabel, pChild->mChildQty, pGrandChildren);
            CNode*      pNewNode = NodeAllocator::mpNodePtr[SubTreeIndex];
    
            // Transpose the first two characters
            std::swap(pNewNode->mpLabel[0], pNewNode->mpLabel[1]);

            CCompTrie<NodeAllocator>    SubTree = CreateTrieWithOneChild(SubTreeIndex);

            FirstTrie = FirstTrie.MergeTries(CCompTrie<NodeAllocator>(SubTree), true /* Reuse nodes, don't copy */);
        } else if (ChildLabelLen) {
        // If label is long enough, we have to concatenate two labels and only then transpose characters
        // This operation is performed for every grand child

            for (SmallCounter GrandChildId = 0; GrandChildId < pChild->mChildQty; ++GrandChildId) {
                CNode*  pGrandChild = NodeAllocator::mpNodePtr[pGrandChildren[GrandChildId]];

                SmallCounter    GrandChildLabelLen = pGrandChild->mLabelLen;

                if (GrandChildLabelLen) {
                    unsigned ConcatLabelLen = GrandChildLabelLen + ChildLabelLen;

                    if (ConcatLabelLen > MAX_SMALL_COUNTER) {
                        THROW_AND_PRINT_RUNTIME_ERROR("The total and child labels sum up to > MAX_SMALL_COUNTER")
                    }

                    char aConcatLabel[ConcatLabelLen];


                    memcpy(aConcatLabel, pChild->mpLabel, ChildLabelLen);
                    memcpy(aConcatLabel + ChildLabelLen, pGrandChild->mpLabel, GrandChildLabelLen);

                    // Transpose the first two characters
                    std::swap(aConcatLabel[0], aConcatLabel[1]);

                    NodeIndex   SubTreeIndex = NodeAllocator::AllocNode(ConcatLabelLen, aConcatLabel, pGrandChild->mChildQty, pGrandChild->GetChildData());
                    CNode*      pNewNode = NodeAllocator::mpNodePtr[SubTreeIndex];

                    CCompTrie<NodeAllocator>    SubTree = CreateTrieWithOneChild(SubTreeIndex);

                    FirstTrie = FirstTrie.MergeTries(CCompTrie<NodeAllocator>(SubTree), true /* Reuse nodes, don't copy */);
                }
            }
        }
    }

    return FirstTrie.mRootIndex;
}

template <typename NodeAllocator>
NodeIndex
CCompTrie<NodeAllocator>::CreateDeletionTrie(NodeIndex StartNode, SmallCounter Centroid)
{
    if (StartNode >= NodeAllocator::mpNodePtr.size()) {
        THROW_AND_PRINT_RUNTIME_ERROR("Invalid node index " << StartNode);
    }
    CNode*          pOldNode = NodeAllocator::mpNodePtr[StartNode];
    NodeIndex*      pChildren = pOldNode->GetChildData();

    /* 
     *  We create a new trie that includes all child non-terminal nodes.
     *  All such children must have non-empty labels.  All labels must be shortened by one char. 
     */

    SmallCounter    NewChildCnt = 0;

    for (SmallCounter i = 0; i < pOldNode->mChildQty; ++i) 
    if (i != Centroid) {
        CNode*      pChild = NodeAllocator::mpNodePtr[pChildren[i]];

        if (pChild->mLabelLen) {
            ++NewChildCnt;
        }
    }

    if (!NewChildCnt) {
        return NILL_NODE_INDEX; // Empty trie
    }

    NodeIndex    CreatedNodes[NewChildCnt];
    SmallCounter CreatedCnt = 0;

    CCompTrie<NodeAllocator> FirstTrie ; // It MUST BE AN EMPTY trie, i.e. no children at all
   
    for (SmallCounter i = 0, k = 0; i < pOldNode->mChildQty; ++i) 
    if (i != Centroid) {
        CNode*                      pChild = NodeAllocator::mpNodePtr[pChildren[i]];

        if (pChild->mLabelLen) {
            CCompTrie<NodeAllocator>    SubTree;

            // Delete the first label character
            NodeIndex   SubTreeIndex = NodeAllocator::AllocNode(pChild->mLabelLen- 1, pChild->mpLabel + 1, pChild->mChildQty, pChild->GetChildData());

            // We ensure that a newly created subtree has a root with empty label
            if (pChild->mLabelLen > 1) {
                SubTree = CreateTrieWithOneChild(SubTreeIndex);
            } else {
                SubTree = SubTreeIndex;
            }

            CreatedNodes[CreatedCnt++] = SubTree.GetRoot();
            FirstTrie = FirstTrie.MergeTries(CCompTrie<NodeAllocator>(SubTree), true /* Reuse nodes, don't copy */);
        }
    }
  
    // We created several root nodes at these point. Let's deallocate unused ones.
    
    for (SmallCounter  i = 0; i < CreatedCnt; ++i) 
    if (FirstTrie.GetRoot() != CreatedNodes[i]) {
        NodeAllocator::FreeNode(CreatedNodes[i]);
    }
    
    return FirstTrie.GetRoot();
}

template <typename NodeAllocator>
NodeIndex
CCompTrie<NodeAllocator>::CreateSubstitutionTrie(NodeIndex StartNode, SmallCounter Centroid)
{
    if (StartNode >= NodeAllocator::mpNodePtr.size()) {
        THROW_AND_PRINT_RUNTIME_ERROR("Invalid node index " << StartNode);
    }
    CNode*          pOldNode = NodeAllocator::mpNodePtr[StartNode];
    NodeIndex*      pChildren = pOldNode->GetChildData();

    /* 
     *  We create a new trie that includes all non-centroid child non-terminal nodes.
     *  All such children must have non-empty labels. 
     * 
     */

    SmallCounter    NewChildCnt = 0;

    for (SmallCounter i = 0; i < pOldNode->mChildQty; ++i) 
    if (i != Centroid) {
        CNode*      pChild = NodeAllocator::mpNodePtr[pChildren[i]];

        if (pChild->mLabelLen) {
            ++NewChildCnt;
        }
    }

    if (!NewChildCnt) {
        return NILL_NODE_INDEX; // Empty trie
    }

    NodeIndex    CreatedNodes[NewChildCnt];
    SmallCounter CreatedCnt = 0;

    CCompTrie<NodeAllocator> FirstTrie ; // It MUST BE AN EMPTY trie, i.e. no children at all
   
    for (SmallCounter i = 0, k = 0; i < pOldNode->mChildQty; ++i) 
    if (i != Centroid) {
        CNode*                      pChild = NodeAllocator::mpNodePtr[pChildren[i]];

        if (pChild->mLabelLen) {
            CCompTrie<NodeAllocator>    SubTree;

            // Substitute the first label character
            NodeIndex   SubTreeIndex = NodeAllocator::AllocNode(pChild->mLabelLen, pChild->mpLabel, pChild->mChildQty, pChild->GetChildData());
            CNode*                      pNewNode = NodeAllocator::mpNodePtr[SubTreeIndex];
    
            pNewNode->mpLabel[0] = WILDCHARD_CHAR;
            SubTree = CreateTrieWithOneChild(SubTreeIndex);
    
            CreatedNodes[CreatedCnt++] = SubTree.GetRoot();
            FirstTrie = FirstTrie.MergeTries(CCompTrie<NodeAllocator>(SubTree), true /* Reuse nodes, don't copy */);
        }
    }
  
    // We created several root nodes at these point. Let's deallocate unused ones.
    
    for (SmallCounter  i = 0; i < CreatedCnt; ++i) 
    if (FirstTrie.GetRoot() != CreatedNodes[i]) {
        NodeAllocator::FreeNode(CreatedNodes[i]);
    }
    
    return FirstTrie.GetRoot();
}

template <typename NodeAllocator>
NodeIndex
CCompTrie<NodeAllocator>::CreateInsertionTrie(NodeIndex StartNode, SmallCounter Centroid)
{
    if (StartNode >= NodeAllocator::mpNodePtr.size()) {
        THROW_AND_PRINT_RUNTIME_ERROR("Invalid node index " << StartNode);
    }
    CNode*          pOldNode = NodeAllocator::mpNodePtr[StartNode];
    NodeIndex*      pChildren = pOldNode->GetChildData();

    /* 
     *  We create a new node that includes all child (non-centroid) nodes,
     *  including terminal nodes with empty labels. 
     * 
     */
    if (pOldNode->mChildQty < 2) {
        return NILL_NODE_INDEX; // empty trie
    }

    SmallCounter    NewChildCnt = pOldNode->mChildQty - 1; // Minus centroid child

    NodeIndex    CreatedNodes[NewChildCnt];
    SmallCounter CreatedCnt = 0;

    CCompTrie<NodeAllocator> FirstTrie ; // It MUST BE AN EMPTY trie, i.e. no children at all
   
    for (SmallCounter i = 0, k = 0; i < pOldNode->mChildQty; ++i) 
    if (i != Centroid) {
        CNode*                      pChild = NodeAllocator::mpNodePtr[pChildren[i]];

        CCompTrie<NodeAllocator>    SubTree;

        // Substitute the first label character
        NodeIndex   SubTreeIndex = NodeAllocator::AllocNode(pChild->mLabelLen + 1, NULL, pChild->mChildQty, pChild->GetChildData());
        CNode*                      pNewNode = NodeAllocator::mpNodePtr[SubTreeIndex];
    
        memcpy(pNewNode->mpLabel + 1, pChild->mpLabel, pChild->mLabelLen);
        pNewNode->mpLabel[0] = WILDCHARD_CHAR;
        SubTree = CreateTrieWithOneChild(SubTreeIndex);
    
        CreatedNodes[CreatedCnt++] = SubTree.GetRoot();
        FirstTrie = FirstTrie.MergeTries(CCompTrie<NodeAllocator>(SubTree), true /* Reuse nodes, don't copy */);
    }
  
    // We created several root nodes at these point. Let's deallocate unused ones.
    
    for (SmallCounter  i = 0; i < CreatedCnt; ++i) 
    if (FirstTrie.GetRoot() != CreatedNodes[i]) {
        NodeAllocator::FreeNode(CreatedNodes[i]);
    }
    return FirstTrie.GetRoot();
}

template <typename NodeAllocator>
NodeIndex
CCompTrie<NodeAllocator>::CreateTrieWithOneChild(NodeIndex OneChild)
{

    NodeIndex    NewRootIndex = NodeAllocator::AllocNode(0, NULL, 1); // A root node
    NodeIndex*   pNewRootChildren = NodeAllocator::mpNodePtr[NewRootIndex]->GetChildData();

    pNewRootChildren[0] = OneChild;

    return NewRootIndex;
}

template <typename NodeAllocator>
NodeIndex
CCompTrie<NodeAllocator>::CreateTrieWithNoChild()
{

    NodeIndex    NewRootIndex = NodeAllocator::AllocNode(0, NULL, 0); // A root node

    return NewRootIndex;
}

template <typename NodeAllocator>
NodeIndex
CCompTrie<NodeAllocator>::Clone(NodeIndex StartNode, bool bReuseExistingNodes)
{
    if (StartNode == NILL_NODE_INDEX) {
        return NILL_NODE_INDEX; // Don't clone the empty trie
    }
    if (bReuseExistingNodes) {
        return StartNode;
    }

    if (StartNode >= NodeAllocator::mpNodePtr.size()) {
        THROW_AND_PRINT_RUNTIME_ERROR("Invalid node index " << StartNode);
    }
    CNode*          pNode = NodeAllocator::mpNodePtr[StartNode];

    NodeIndex       pNewChildren[pNode->mChildQty];
    NodeIndex*      pOldChildren = pNode->GetChildData();

    for (SmallCounter i = 0; i < pNode->mChildQty; ++i) {
        NodeIndex ClonedChildIndex = Clone(pOldChildren[i], false); // Actually copy, do not reuse existing nodes

        pNewChildren[i] = ClonedChildIndex;
    }

    return NodeAllocator::AllocNode(pNode->mLabelLen, pNode->mpLabel, pNode->mChildQty, pNewChildren);
};

template <typename NodeAllocator> 
void 
CCompTrie<NodeAllocator>::AccessNode(NodeIndex Index, SmallCounter& LabelLen, const char*& pLabel, SmallCounter& ChildQty, NodeIndex*& pChildren)
{

    CNode*          pNode = (Index < NodeAllocator::mpNodePtr.size()) ? NodeAllocator::mpNodePtr[Index] : NULL;
    
    if (!pNode) {
        std::stringstream err;

        err << "Invalid node index: " << Index;
        THROW_AND_PRINT_RUNTIME_ERROR(err.str());
    }

    pLabel    = pNode->mpLabel;
    ChildQty  = pNode->mChildQty;
    pChildren = pNode->GetChildData();
    LabelLen  = pNode->mLabelLen;
}

template <typename NodeAllocator> 
void 
CCompTrie<NodeAllocator>::AccessNode(NodeIndex Index, SmallCounter& LabelLen, const char*& pLabel, SmallCounter& ChildQty, const NodeIndex*& pChildren) const
{

    CNode*          pNode = (Index < NodeAllocator::mpNodePtr.size()) ? NodeAllocator::mpNodePtr[Index] : NULL;
    
    if (!pNode) {
        std::stringstream err;

        err << "Invalid node index: " << Index;
        THROW_AND_PRINT_RUNTIME_ERROR(err.str());
    }

    pLabel    = pNode->mpLabel;
    ChildQty  = pNode->mChildQty;
    pChildren = pNode->GetChildData();
    LabelLen  = pNode->mLabelLen;
}

template <typename NodeAllocator> 
template <typename Visitor>
void CCompTrie<NodeAllocator>::VisitTrie(Visitor &vis) const
{
    if (Empty()) {
        return;
    }

    std::stack<CEntry>  q;

    q.push(CEntry(mRootIndex));

    while (!q.empty()) {
        NodeIndex       CurrIndex = q.top().mCurrIndex;
        unsigned int    CurrLevel = q.top().mLevel;
        SmallCounter    CurrChild = q.top().mCurrChild;
        CNode*          pCurrNode = NodeAllocator::mpNodePtr[CurrIndex];
        SmallCounter    ChildQty  = pCurrNode->mChildQty;

        q.pop();

        if (!CurrChild) {
            vis(CurrIndex, CurrLevel, pCurrNode->mLabelLen, pCurrNode->mpLabel, ChildQty);
        }

        if (CurrChild < ChildQty) {
            NodeIndex*  pChildren = pCurrNode->GetChildData();

            assert(pChildren[CurrChild] < NodeAllocator::mpNodePtr.size());
            q.push(CEntry(CurrIndex, CurrChild +1, CurrLevel));
            q.push(CEntry(pChildren[CurrChild], 0, CurrLevel + 1));
        }
    }

}

template <typename NodeAllocator> 
void CCompTrie<NodeAllocator>::GetNodeInfo(NodeIndex Index, SmallCounter &ChildQty, SmallCounter &LabelLen, const char*& pLabel, const NodeIndex*& pChildren) const
{
    if (Index >= NodeAllocator::mpNodePtr.size()) {
        std::stringstream err;
        err << "Node index " << Index << " is out of bounds";
        THROW_AND_PRINT_RUNTIME_ERROR(err.str());
    }

    CNode*      pNode = NodeAllocator::mpNodePtr[Index];

    ChildQty    = pNode->mChildQty;
    pChildren   = pNode->GetChildData();
    LabelLen    = pNode->mLabelLen;
    pLabel      = pNode->mpLabel;
}

template <typename NodeAllocator> 
void CCompTrie<NodeAllocator>::Insert(const std::string& Str)
{
    if (Empty()) {
        mRootIndex = NodeAllocator::AllocNode(0, NULL, 1); // A root node
        NodeIndex*  pChildren = NodeAllocator::mpNodePtr[mRootIndex]->GetChildData();

        pChildren[0] = NodeAllocator::AllocNode(Str.size(), Str.c_str(), 0);
        return;
    }

    NodeIndex       CurrNodeIndex = mRootIndex;
    CNode*          pCurrNode = NodeAllocator::mpNodePtr[CurrNodeIndex];

    SmallCounter    Start     = 0;

    while (1) {
        SmallCounter    ChildNum  = 0;
        SmallCounter    MaxLen    = 0;

        NodeIndex*      pChildren = pCurrNode->GetChildData();

        if (!pChildren) {
            const std::string&  Label1       = Str.substr(Start);
            NodeIndex           TerminalNode = NodeAllocator::AllocNode(0, NULL, 0); // Just a terminal node with empty label
            NodeIndex           NewStrNode   = NodeAllocator::AllocNode(static_cast<SmallCounter>(Label1.size()), Label1.c_str(), 0);
            SmallCounter        OldLabelLen  = pCurrNode->mLabelLen;

            assert(Start >= OldLabelLen && 
                   Str.substr(Start - OldLabelLen, OldLabelLen) == std::string(pCurrNode->mpLabel, OldLabelLen));

            NodeAllocator::ReallocNode(CurrNodeIndex, OldLabelLen, 2);

            // Now reassign pointers
            pCurrNode = NodeAllocator::mpNodePtr[CurrNodeIndex];
            pChildren = pCurrNode->GetChildData();

            pChildren[0] = TerminalNode; // Terminal node always has an empty label, which is lexicographically smaller than any non-empty one.
            pChildren[1] = NewStrNode;

            break;
        }

        assert(pChildren);
        assert(pCurrNode->mChildQty);

        for (SmallCounter i = 0; i < pCurrNode->mChildQty; ++i) {
            SmallCounter len = CheckCommon(pChildren[i], Str, Start);

            if (len > MaxLen) {
                ChildNum = i;
                MaxLen = len;
            }
        }

        if (!MaxLen) {
            // We need to create one new node and allocate more space for the current one
            const std::string&  Label   = Str.substr(Start);
            NodeIndex           NewNode = NodeAllocator::AllocNode(static_cast<SmallCounter>(Label.size()), Label.c_str(), 0);
            SmallCounter        OldChildQty = pCurrNode->mChildQty;

            NodeAllocator::ReallocNode(CurrNodeIndex, pCurrNode->mLabelLen, OldChildQty + 1);

            // Now reassign pointers
            pCurrNode = NodeAllocator::mpNodePtr[CurrNodeIndex];
            pChildren = pCurrNode->GetChildData();

            // Should find a position to insert a new child:
            // We want to keep child labels in alphabetical order.
            SmallCounter InsertPos = 0;

            for (; InsertPos < OldChildQty; ++InsertPos) {
                const CNode* pChild = NodeAllocator::mpNodePtr[pChildren[InsertPos]]; 

                if (strncmp(Label.data(),  pChild->mpLabel, pChild->mLabelLen) < 0) {
                    break;
                }
            }

            if (InsertPos < OldChildQty) {
                memmove(pChildren + InsertPos + 1, pChildren + InsertPos, sizeof(*pChildren) * (OldChildQty - InsertPos)); 
            }
            pChildren[InsertPos] = NewNode;

            break;
        }

        assert(MaxLen);

        NodeIndex  ChildIndex = pChildren[ChildNum]; 
        CNode*     pChild     = NodeAllocator::mpNodePtr[ChildIndex];

        if (MaxLen == pChild->mLabelLen) { // Insert into a child
            Start += MaxLen;
            pCurrNode = pChild;
            CurrNodeIndex = ChildIndex;
            continue;
        }

        assert(MaxLen < pChild->mLabelLen); 

        // Introduce two new nodes, shorter then label of child with #ChildNum
        const std::string&  LabelNewStr = Str.substr(Start + MaxLen);

        SmallCounter        Label2Len = pChild->mLabelLen - MaxLen;
        const std::string   Label2(pChild->mpLabel + MaxLen, Label2Len);

        NodeIndex       NewChildNode  = NodeAllocator::AllocNode(MaxLen, pChild->mpLabel, 2);
        NodeIndex       NewStrNode    = NodeAllocator::AllocNode(static_cast<SmallCounter>(LabelNewStr.size()), LabelNewStr.c_str(), 0);

        NodeAllocator::ReallocNode(ChildIndex, Label2Len, pChild->mChildQty); // The child node needs less space now

        // Now reassign the child pointer & modify label
        pChild = NodeAllocator::mpNodePtr[ChildIndex];
        memcpy(pChild->mpLabel, Label2.c_str(), Label2.size());

        NodeIndex* pNewChildren = NodeAllocator::mpNodePtr[NewChildNode]->GetChildData();

        pNewChildren[0] = NewStrNode;
        pNewChildren[1] = ChildIndex;

        if (LabelNewStr > Label2) {
            std::swap(pNewChildren[0], pNewChildren[1]);
        }
        pChildren[ChildNum] = NewChildNode;

        break;
    }
}

template <typename NodeAllocator>
NodeIndex  CCompTrie<NodeAllocator>::MergeTriesInternal(NodeIndex RootIndex1, NodeIndex RootIndex2, bool bReuseExistingNodes) const
{
    // Node labels might be different here.
    // The label of the first node is always used.

    const CNode *pNode1 = NodeAllocator::mpNodePtr[RootIndex1];
    const CNode *pNode2 = NodeAllocator::mpNodePtr[RootIndex2];

    if (!pNode1->mChildQty && !pNode2->mChildQty) {
        return Clone(RootIndex1, bReuseExistingNodes);
    }

    const NodeIndex*  pChildren1 = pNode1->GetChildData();
    const NodeIndex*  pChildren2 = pNode2->GetChildData();
    
    // The case, when one child has children, but another does not, should be treated separately
    // The only case, when node labels may differ here is only when both nodes have children (see Comment ** below).
    if (!pNode2->mChildQty || !pNode1->mChildQty) {
        assert(pNode1->mLabelLen == pNode2->mLabelLen && !strncmp(pNode1->mpLabel, pNode2->mpLabel, pNode1->mLabelLen));
        if (pNode1->mChildQty) {
            std::swap(pNode1, pNode2);
            std::swap(RootIndex1, RootIndex2);
            std::swap(pChildren1, pChildren2);
        }
        // If one of the pNode2 child is a terminal node with the empty label, we just return RootIndex2
        CNode       *pChild2 = NodeAllocator::mpNodePtr[pChildren2[0]];
        
        if (!pChild2->mLabelLen) {
            assert(!pChild2->mChildQty);
            return Clone(RootIndex2, bReuseExistingNodes); 
        }
        // Here we do not have children with empty label.
        
        // In this case, we create a joint node that contains a terminal child with the empty label and children of node pNode2 (see swap code)
        NodeIndex   NewRootIndex = NodeAllocator::AllocNode(pNode1->mLabelLen, pNode1->mpLabel, pNode2->mChildQty + 1);
        CNode*      pNewNode = NodeAllocator::mpNodePtr[NewRootIndex];
        NodeIndex*  pNewChildren = pNewNode->GetChildData();
        
        pNewChildren[0] =  NodeAllocator::AllocNode(0, NULL, 0);
        
        /* If no cloning just do: memcpy(&pNewChildren[1], pChildren2, pNode2->mChildQty * sizeof(NodeIndex)); */
        
        for (SmallCounter i = 1; i <= pNode2->mChildQty; ++i) {
            pNewChildren[i] = Clone(pChildren2[i - 1], bReuseExistingNodes);
        }

        return NewRootIndex;
    }

    int i1 = 0, i2 = 0;

    int NewChildQty = 0;
    int MergeMap[MAX_SMALL_COUNTER][3];

    std::fill(MergeMap[0], MergeMap[0] + sizeof MergeMap / sizeof **MergeMap, -1);

    for (;i1 < pNode1->mChildQty && i2 < pNode2->mChildQty;++NewChildQty) {
        CNode*  p1 = NodeAllocator::mpNodePtr[pChildren1[i1]];
        CNode*  p2 = NodeAllocator::mpNodePtr[pChildren2[i2]];

        int     PrefLen = CheckCommon(p1->mpLabel, p1->mLabelLen, 
                                      p2->mpLabel, p2->mLabelLen);

        if (PrefLen || (!p1->mLabelLen && !p2->mLabelLen)) {
            // Child i1 and i2 gonna be merged
            MergeMap[NewChildQty][0] = i1++;
            MergeMap[NewChildQty][1] = i2++;
            MergeMap[NewChildQty][2] = PrefLen;

            continue;
        }
        int cmp = strncmp(p1->mpLabel, p2->mpLabel, std::min(p1->mLabelLen, p2->mLabelLen));
        assert(bool(p1->mLabelLen) || bool(p2->mLabelLen));
        if (!p1->mLabelLen) { cmp = -1; }
        if (!p2->mLabelLen) { cmp = +1; }

        if (cmp < 0) {
            MergeMap[NewChildQty][0] = i1++;
        } else {
            MergeMap[NewChildQty][1] = i2++;
        }
    }

    while (i1 < pNode1->mChildQty) {
        MergeMap[NewChildQty++][0] = i1++;
    }
    while (i2 < pNode2->mChildQty) {
        MergeMap[NewChildQty++][1] = i2++;
    }

    // Always use the label of the first node here
    NodeIndex   NewRootIndex = NodeAllocator::AllocNode(pNode1->mLabelLen, pNode1->mpLabel, NewChildQty);
    CNode*      pNewRoot     = NodeAllocator::mpNodePtr[NewRootIndex];
    NodeIndex*  pNewChildren = pNewRoot->GetChildData(); 

    for (int i = 0; i < NewChildQty; ++i) {
        int Child1 = MergeMap[i][0];
        int Child2 = MergeMap[i][1];
        int PrefLen = MergeMap[i][2];

        assert(Child1 >= 0 || Child2 >= 0);
        if (Child1 < 0) {
            pNewChildren[i] = pChildren2[Child2];
        } else if (Child2 < 0) {
            pNewChildren[i] = pChildren1[Child1];
        } else {
            NodeIndex   Index1 = pChildren1[Child1];
            NodeIndex   Index2 = pChildren2[Child2];
            CNode*      p1 = NodeAllocator::mpNodePtr[Index1];
            CNode*      p2 = NodeAllocator::mpNodePtr[Index2];

            SmallCounter    Label1Len = p1->mLabelLen - PrefLen;
            SmallCounter    Label2Len = p2->mLabelLen - PrefLen;

            std::string     Label1Rest(p1->mpLabel + PrefLen, Label1Len);
            std::string     Label2Rest(p2->mpLabel + PrefLen, Label2Len);

            std::string     CommLabel(p1->mpLabel, PrefLen);

            if (Label1Rest > Label2Rest) {
                std::swap(p1, p2);
                std::swap(Label1Rest, Label2Rest);
                std::swap(Label1Len, Label2Len);
                std::swap(Index1, Index2);
            }

            //
            // Three cases are possible:
            //      1. Labels are equal and both nodes have children (doing recursion)
            //      2. One label is a prefix of another label. 
            //         Index2 has a longer label (see swapping code above), while 
            //         Index1 has no children.
            //         Then we create a new node Index2Mode that is a copy of Index2, except
            //         its label is Label2Rest.  Then we create a parent for Index2Mod with label CommLabel,
            //         whose only child is Index2Mod.  Its index is named Index2ModParent.
            //         Merge Index2ModParent and Index1.
            //         Dispose, Index2ModParent, because it is not incorporated into the merged tree,
            //         see the description of case 3.
            //         This trick is necessary to avoid creation of nodes with EMPTY LABELS.
            //      3. None of the labels is a proper prefix of another label. In this
            //         case we create an intermediate node with 2 children: Index1Mod and Index2Mod.
            //         These nodes are copies of Index1 and Index2, respectively, except their labels
            //         are Label1Rest and Label2Rest.
            //
            if (!Label1Len && !Label2Len) {
                /* 
                 * CASE 1.
                 */
                pNewChildren[i] = MergeTriesInternal(Index1, Index2, bReuseExistingNodes);
            } else if ((!Label1Len && p1->mChildQty) || (!Label2Len && p2->mChildQty)) {
                /*
                 * CASE 2.
                 */
                
                // A copy of Index2, but with a modified label
                /* If no cloning just do:
                NodeIndex Index2Mod = NodeAllocator::AllocNode(Label2Len, Label2Rest.c_str(), p2->mChildQty, p2->GetChildData());
                */
                NodeIndex   pNewChild2[p2->mChildQty];
                NodeIndex*  pOldChild2 = p2->GetChildData();

                for (SmallCounter k = 0; k < p2->mChildQty; ++k) {
                    pNewChild2[k] = Clone(pOldChild2[k], bReuseExistingNodes);
                }

                NodeIndex Index2Mod = NodeAllocator::AllocNode(Label2Len, Label2Rest.c_str(), p2->mChildQty, pNewChild2);

                CNode*    p2Mod = NodeAllocator::mpNodePtr[Index2Mod];

                // Create a new root node, whose only child is Index2Mod
                NodeIndex   Index2ModParent = NodeAllocator::AllocNode(PrefLen, CommLabel.c_str(), 1);
                NodeIndex*  pTmpChildren = NodeAllocator::mpNodePtr[Index2ModParent]->GetChildData();
                pTmpChildren[0] = Index2Mod;


                // Comment **
                // We do not have to create a modified copy of Index1 with label Label1Rest, because:
                //      1. The number of children is not gonna change;
                //      2. The label will be ignored by recursive call of MergeTriesInternal.
                
                // Merge Trees. Function uses the label of the node that corresponds to the first index.
                pNewChildren[i] = MergeTriesInternal(Index2ModParent, Index1, bReuseExistingNodes); 

                // We can discard Index2ModParent, since it is not used in the merge
                assert(pNewChildren[i] != Index2ModParent);
                NodeAllocator::FreeNode(Index2ModParent);
            } else {
                /*
                 * CASE 3.
                 */
                NodeIndex Index1Mod = NodeAllocator::AllocNode(Label1Len, Label1Rest.c_str(), p1->mChildQty, p1->GetChildData());
                CNode*    p1Mod = NodeAllocator::mpNodePtr[Index1Mod];

                NodeIndex Index2Mod = NodeAllocator::AllocNode(Label2Len, Label2Rest.c_str(), p2->mChildQty, p2->GetChildData());
                CNode*    p2Mod = NodeAllocator::mpNodePtr[Index2Mod];

                NodeIndex   IndexInterm = NodeAllocator::AllocNode(PrefLen, CommLabel.c_str(), 2);
                NodeIndex*  pTmpChildren = NodeAllocator::mpNodePtr[IndexInterm]->GetChildData();

                // Labels Index1Mod and Index2Mod are already reasorted, see swapping code above
                pTmpChildren[0] = Index1Mod;
                pTmpChildren[1] = Index2Mod;

                assert(Label1Len == Label1Rest.size());
                assert(Label2Len == Label2Rest.size());

                pNewChildren[i] = IndexInterm;
            }
        }
    }

    return NewRootIndex;
}

template <typename NodeAllocator>
CCompTrie<NodeAllocator> CCompTrie<NodeAllocator>::MergeTries(const CCompTrie &OthTree, bool bReuseExistingNodes) const
{
    if (Empty() && OthTree.Empty()) {
        return CCompTrie<NodeAllocator>(NILL_NODE_INDEX);
    }
    if (Empty()) {
       return CCompTrie(Clone(OthTree.GetRoot(), bReuseExistingNodes)); 
    }
    if (OthTree.Empty()) {
        return CCompTrie(Clone(this->GetRoot(), bReuseExistingNodes));
    }

    return CCompTrie(MergeTriesInternal(mRootIndex, OthTree.mRootIndex, bReuseExistingNodes));
}

template <typename NodeAllocator> 
SmallCounter  CCompTrie<NodeAllocator>::CheckCommon(const char *pStr1, SmallCounter Len1, const char *pStr2, SmallCounter Len2) const
{
    SmallCounter len = std::min(Len1, Len2);

    int res = 0;
    for (; res < len; ++res) 
    if (pStr1[res] != pStr2[res]) { break; }

    return res;
}

template <typename NodeAllocator> 
SmallCounter CCompTrie<NodeAllocator>::CheckCommon(NodeIndex Index, const std::string &Str, SmallCounter Start) const
{
    CNode*    pNode = NodeAllocator::mpNodePtr[Index];

    SmallCounter res = 0; 
    SmallCounter len = std::min(pNode->mLabelLen, static_cast<SmallCounter>(Str.size() - Start));
    const char*  pLabel = pNode->mpLabel;

    for (res = 0; res < len; ++res) 
    if (pLabel[res] != Str[Start++]) { break; }

    return res; 
}

#endif
