#include <algorithm>
#include <set>

#include "testcomptrie.h"
#include "testutil.h"
#include "../Common/mytest.h"

using namespace std;

typedef vector<string>              vs;
typedef set<string>                 ss;
typedef CCompTrie<CNodeAllocator>   TTrie;

void PrintTries(const TTrie &trie1, const TTrie &trie2, const TTrie &trie3) {

    PrintTrie(trie1);
    PrintTrieStrings(trie1);
    cerr << "======================\n";
    PrintTrie(trie2);
    PrintTrieStrings(trie2);
    cerr << "======================\n";

    cerr << "== Merged ============\n";

    PrintTrie(trie3);
    PrintTrieStrings(trie3);

    cerr << "======================\n";

}

bool CheckTrieStrings(const TTrie& trie, vs SourceList) {
    vs      TrieList;

    sort(SourceList.begin(), SourceList.end());

    GetTrieStrings(trie, TrieList);

    // The list must be sorted
    for (int i = 1; i < TrieList.size(); ++i) {
        if (TrieList[i] <= TrieList[i - 1]) {
            cerr << "Comparison failed for strings: '" << TrieList[i-1] << "' and '" << TrieList[i] << "'" << endl;
            return false;
        }
    }

    
    vs Diff;
    set_difference(SourceList.begin(), SourceList.end(), TrieList.begin(), TrieList.end(), back_inserter(Diff));

    if (Diff.size()) {
        cerr << "The following is the difference between string sets:" << endl;
        for (int i = 0; i < Diff.size(); ++i) {
            cerr << "'" << Diff[i] << "'" << endl;
        }
        cerr << "=========================" << endl;
        return false;
    }

    return true;
}

bool DoCommonOpCheck(const vs& list1, const vs& list2, bool bReuseExistingNodes) {
    TTrie trie1, trie2, trie3;

    // Check the first trie
    size_t Mem1 = CNodeAllocator::GetMemUsedByNodes();
    size_t Qty1 = CNodeAllocator::GetNodeQty();

    for (int i = 0; i < list1.size(); ++i) {
        trie1.Insert(list1[i]);
    }
    
    size_t Mem2 = CNodeAllocator::GetMemUsedByNodes();
    size_t Qty2 = CNodeAllocator::GetNodeQty();

    size_t UsedMem, NodeQty; 
    
    bool check1  = CheckTrie(trie1, UsedMem, NodeQty);
    bool check11 = CheckTrieStrings(trie1, list1);

    if (UsedMem != Mem2 - Mem1 || NodeQty != Qty2 - Qty1) {
        PrintTrie(trie1);
        PrintTrieStrings(trie1);
        cerr << "======================\n";
    }

    TEST_COND_MSG_RETURN_FALSE(UsedMem == Mem2 - Mem1, "UsedMem != Mem2 - Mem1");
    TEST_COND_MSG_RETURN_FALSE(NodeQty == Qty2 - Qty1, "NodeQty != Qty2 - Qty1");

    // Check the second trie

    for (int i = 0; i < list2.size(); ++i) {
        trie2.Insert(list2[i]);
    }

    bool check2 = CheckTrie(trie2, UsedMem, NodeQty);
    bool check21 = CheckTrieStrings(trie2, list2);

    size_t Mem3 = CNodeAllocator::GetMemUsedByNodes();
    size_t Qty3 = CNodeAllocator::GetNodeQty();

    TEST_COND_MSG_RETURN_FALSE(UsedMem == Mem3 - Mem2, "UsedMem != Mem3 - Mem2");
    TEST_COND_MSG_RETURN_FALSE(NodeQty == Qty3 - Qty2, "NodeQty != Qty3 - Qty2");


    // Check the interesection (no memory check now, because trie3 can share nodes with trie1 and/or trie2

    trie3 = trie1.MergeTries(trie2, bReuseExistingNodes);

    vs  ListMerged;

    set_union(list1.begin(), list1.end(), list2.begin(), list2.end(), back_inserter(ListMerged));

    bool check3 = CheckTrie(trie3, UsedMem, NodeQty);
    bool check31 = CheckTrieStrings(trie3, ListMerged);

    /* 
     * Now check how cloning works:
     *  1) We want to ensure that a cloned tree generates the same subset of strings as an original one.
     *  2) If cloning is done without sharing nodes, i.e. true copy on read, we ensure that no nodes
     *     are shared between two trees.
     */

    set<NodeIndex>      trie3Nodes, trie3NodesClone;
    set<NodeIndex>      CloneInter;

    TTrie   trie3Clone = trie3.CloneTrie();

    bool check32 = CheckTrieStrings(trie3Clone, ListMerged);

    GetTrieNodes(trie3, trie3Nodes);
    GetTrieNodes(trie3Clone, trie3NodesClone);

    set_intersection(trie3Nodes.begin(), trie3Nodes.end(), trie3NodesClone.begin(), trie3NodesClone.end(), inserter(CloneInter, CloneInter.begin()));

    bool check33 = CloneInter.empty() && trie3Nodes.size() == trie3NodesClone.size();

    if (!check1 || !check2 || !check3 || !check11 || !check21 || !check31) {
        PrintTries(trie1, trie2, trie3);
        PrintTrie(trie3Clone);
    }

    if (!check32 || !check33) {
        cerr << "Intersection size: " << CloneInter.size() << " Orig trie node set size: " << trie3Nodes.size() << " Cloned trie node set size: " << trie3NodesClone.size() << endl; 
        PrintTrie(trie3);
        PrintTrie(trie3Clone);
    }

    TEST_COND_MSG_RETURN_FALSE(check1, "check1");
    TEST_COND_MSG_RETURN_FALSE(check2, "check2");
    TEST_COND_MSG_RETURN_FALSE(check3, "check3");
    TEST_COND_MSG_RETURN_FALSE(check11, "check11");
    TEST_COND_MSG_RETURN_FALSE(check21, "check21");
    TEST_COND_MSG_RETURN_FALSE(check31, "check31");
    TEST_COND_MSG_RETURN_FALSE(check32, "check32");
    TEST_COND_MSG_RETURN_FALSE(check33, "check33");

    return true;
}

bool test1(bool bReuseExistingNodes)
{

    vs trie1, trie2;


    trie1.push_back("dd");
    trie1.push_back("aa");
    trie1.push_back("a");
    trie1.push_back("ab");
    trie1.push_back("abtest");
    trie1.push_back("ac");
    trie1.push_back("acdd");

    trie2.push_back("aac");
    trie2.push_back("aab");

    if (DoCommonOpCheck(trie1, trie2, bReuseExistingNodes)) {
        return false;
    }
    
    return true;
}

bool test2(bool bReuseExistingNodes)
{

    vs trie1, trie2;


    trie1.push_back("a");
    trie1.push_back("b");
    trie1.push_back("c");

    trie2.push_back("a");
    trie2.push_back("aa");
    trie2.push_back("aaa");
    
    if (!DoCommonOpCheck(trie1, trie2, bReuseExistingNodes)) {
        return false;
    }
    
    return true;
}

bool test3(bool bReuseExistingNodes)
{

    vs trie1, trie2;


    trie1.push_back("a");

    trie2.push_back("aa");

    if (!DoCommonOpCheck(trie1, trie2, bReuseExistingNodes)) {
        return false;
    }
    
    return true;
}

bool test4(bool bReuseExistingNodes)
{

    vs trie1, trie2;


    trie1.push_back("a");

    if (!DoCommonOpCheck(trie1, trie2, bReuseExistingNodes)) {
        return false;
    }
    
    return true;
}

bool test5(bool bReuseExistingNodes)
{

    vs trie1, trie2;

    trie2.push_back("a");

    if (!DoCommonOpCheck(trie1, trie2, bReuseExistingNodes)) {
        return false;
    }
    
    return true;
}

bool test6(bool bReuseExistingNodes)
{

    vs trie1, trie2;

    trie1.push_back("z");

    trie2.push_back("zz");

    if (!DoCommonOpCheck(trie1, trie2, bReuseExistingNodes)) {
        return false;
    }
    
    return true;
}

bool test7(bool bReuseExistingNodes)
{

    vs trie1, trie2;

    trie1.push_back("zz");

    trie2.push_back("z");

    if (!DoCommonOpCheck(trie1, trie2, bReuseExistingNodes)) {
        return false;
    }
    
    return true;
}

bool test8(bool bReuseExistingNodes)
{

    vs trie1, trie2;

    trie1.push_back("a");
    trie1.push_back("aa");

    trie2.push_back("aab");
    trie2.push_back("aac");

    if (!DoCommonOpCheck(trie1, trie2, bReuseExistingNodes)) {
        return false;
    }
    
    return true;
}

bool TestRandom(int MinLen, int MaxLen, int Cnt1, int Cnt2, bool bReuseExistingNodes)
{
    vs trie1, trie2;
    ss strs1, strs2;

    string s;

    while (strs1.size() < Cnt1) {
        GenRandString(MinLen, MaxLen, s);

        strs1.insert(s);
    }

    while (strs2.size() < Cnt2) {
        GenRandString(MinLen, MaxLen, s);

        strs2.insert(s);
    }

    copy(strs1.begin(), strs1.end(), back_inserter(trie1));
    copy(strs2.begin(), strs2.end(), back_inserter(trie2));

    if (!DoCommonOpCheck(trie1, trie2, bReuseExistingNodes)) {
        return false;
    }
    if (!DoCommonOpCheck(trie2, trie1, bReuseExistingNodes)) {
        return false;
    }

    return true;
}
