#include <sys/timeb.h>
#include <stdlib.h>
#include <string.h>

#include <iostream>
#include <vector>
#include <stdexcept>

using namespace std;

#include "int_coder_inl.h"
#include "fileutil.h"

#include "macrotest.h"

const char * file1 = "fileDeltaInl.tmp";
const char * file2 = "fileVarByteInl.tmp";
char       * pMemBuf = NULL;
size_t       memBufSize = 0;

void testInLineWrite(const VI & data)
{
    CFileWrapper file;
    file.open(file1, O_RDWR | O_CREAT | O_TRUNC);

    CBitWriter writer(1024 * 256);

    writer.StartOutput(file.getFd());

    for (unsigned int i = 0; i < data.size(); ++i) {
        DeltaEncode(data[i], writer);
    }

    writer.FinishOutput();

    file.close();

    return;
}

void testInLineRead(const VI & data)
{
    CFileWrapper file;
    file.open(file1, O_RDONLY);

    CBitReader reader(1024 * 256);

    reader.StartInput(file.getFd());

    unsigned int i = 0;
    int      t;

    try {

        for (;i < data.size(); ++i) {
            DeltaDecode(t, reader);
            if (t != data[i]) {
                cerr << "data element # " << i << " does not match, got: " << t << " expected " << data[i] << endl;
                return;
            }
        }
    } catch (const CReachedEof &) {
        cerr << "EOF prematurely ended on line " << i << " of " << data.size() << endl;
        return;
    }

    return;
}

void testInLineVarByteWriteMem(const VI & data)
{
    // First of all we have to re-init global memory buffer
    memBufSize = data.size() * (sizeof(data[0]) + 1); // no more excess byte is need per
    // integer for variable byte encoding
    
    if (pMemBuf) delete [] pMemBuf;
    pMemBuf = new char[memBufSize];

    CMemBitWriter writer(pMemBuf, memBufSize);

    for (unsigned int i = 0; i < data.size(); ++i) {
        try {
            VarByteEncode(data[i], writer);
        } catch (const CEndOfMem & ) {
            cerr << "EOF prematurely ended on line " << i << " of " << data.size() << endl;
            return;
        }
    }

    writer.FinishOutput();

    return;
}

void testInLineVarByteReadMem(const VI & data)
{
    CMemBitReader reader(pMemBuf, memBufSize);

    unsigned int i = 0;
    int      t;

    try {

        for (;i < data.size(); ++i) {
            VarByteDecode(t, reader);
            if (t != data[i]) {
                cerr << "data element # " << i << " does not match, got: " << t << " expected " << data[i] << endl;
                return;
            }
        }
    } catch (const CReachedEof &) {
        cerr << "EOF prematurely ended on line " << i << " of " << data.size() << endl;
        return;
    }

    return;
}

void testInLineVarByteWrite(const VI & data)
{
    CFileWrapper file;
    file.open(file2, O_RDWR | O_CREAT | O_TRUNC);

    CBitWriter writer(1024 * 256);

    writer.StartOutput(file.getFd());

    for (unsigned int i = 0; i < data.size(); ++i) {
        VarByteEncode(data[i], writer);
    }

    writer.FinishOutput();

    file.close();

    return;
}

void testInLineVarByteRead(const VI & data)
{
    CFileWrapper file;
    file.open(file2, O_RDONLY);

    CBitReader reader(1024 * 256);

    reader.StartInput(file.getFd());

    unsigned int i = 0;
    int      t;

    try {

        for (;i < data.size(); ++i) {
            VarByteDecode(t, reader);
            if (t != data[i]) {
                cerr << "data element # " << i << " does not match, got: " << t << " expected " << data[i] << endl;
                return;
            }
        }
    } catch (const CReachedEof &) {
        cerr << "EOF prematurely ended on line " << i << " of " << data.size() << endl;
        return;
    }

    return;
}

void measure(const VI & data, const char * funcName, func funcPtr) 
{
    timeb t1, t2;

    ftime(&t1);
    (*funcPtr)(data);
    ftime(&t2);

    cout << "time for '" << funcName << "' " << (t2.time - t1.time) * 1000 + t2.millitm - t1.millitm << endl;
}

void usage(const char * prog)
{
    cerr << "Usage:\n" << prog << " number of kylobates\n";
}

int main(int argc, char * argv[]) {
    int N = 0;
    if (argc != 2 || !(N = (1024 * atol(argv[1]) / sizeof(int)))) {
        usage(argv[0]);
        return 1;
    }

    try {

       cout << " test to encode/decode " << N << " integers\n";

       VI   data(N); 

       for (unsigned i = 0; i < data.size(); ) {
           data[i] = rand();
           if (!data[i]) {
               continue; // ZERO MUST NOT BE TESTED, it is illegal value to code
           }
           ++i;
       }

       measure(data, "testMacroWrite", testMacroWrite);
       measure(data, "testMacroRead", testMacroRead);
       measure(data, "testInlineWrite", testInLineWrite);
       measure(data, "testInlineRead", testInLineRead);
       measure(data, "testInlineVarByteWrite", testInLineVarByteWrite);
       measure(data, "testInlineVarByteRead", testInLineVarByteRead);
       measure(data, "testInlineVarByteWriteMem", testInLineVarByteWriteMem);
       measure(data, "testInlineVarByteReadMem", testInLineVarByteReadMem);
    } catch (exception & e) {
        cerr << "Exception: " << e.what() << endl;
        return 1;
    }

    return 0;
};
