#ifndef INT_CODER_INL_H
#define INT_CODER_INL_H

#include <sys/types.h>
#include <unistd.h>

#include "except.h"
#include "cmnutil.h"

typedef int32_t int32;
typedef u_int32_t uint32;

struct CReachedEof {
};

struct CEndOfMem {
};
/*
 * Writer is implemented for buffered file read/write
 * and for reading/writing in memory
 */
class CBitWriter {
private:
    size_t         mBufSize;
    size_t         mBytePos;
    char *         mpBuf;
    unsigned char  mBit;
    unsigned char  mChar;
    int            mFd;

    inline void FlushBuf() {
        if (mBytePos && write(mFd, mpBuf, mBytePos) < mBytePos) {
            THROW_RUNTIME_ERROR(SysError());
        }
        mBytePos = 0;
    }
public:
    CBitWriter(size_t bufSize) : mBufSize(bufSize), mpBuf(new char[bufSize]), mFd(-1){}
    inline void StartOutput(int fd) {
        mFd = fd; 
        mBit = 1; 
        mBytePos = 0; 
        mChar = 0;
    }
    inline void PutBit(unsigned int v) {
        if (!mBit) { // After 128 was shifted by 
            mpBuf[mBytePos++] = mChar;
            mBit = 1;
            mChar = 0;
            if (mBytePos >= mBufSize ) {
                FlushBuf();
            }
        }
        if (v) mChar |= mBit;
        mBit <<= 1;    						
    }
    inline void PutByte(unsigned char c) {
        mpBuf[mBytePos++] = c;
        if (mBytePos >= mBufSize ) {
            FlushBuf();
        }
    }
    inline void FinishOutput() {
        if (mBit!=1) {
            mpBuf[mBytePos++]= mChar;
        }
        FlushBuf();
        mBytePos=0;
        mBit=1;
    }
};

class CMemBitWriter {
private:
    size_t         mBufSize;
    size_t         mBytePos;
    char *         mpBuf;
    unsigned char  mBit;
    unsigned char  mChar;

    inline void StartOutput() {
        mBit = 1; 
        mBytePos = 0; 
        mChar = 0;
    }
public:
    CMemBitWriter(char * pBuf, size_t bufSize) : mBufSize(bufSize), mpBuf(pBuf) {
        if (!bufSize) {
            THROW_RUNTIME_ERROR("Zero buffer is not allowed");
        }
        StartOutput();
    }
    inline void PutBit(unsigned int v) {
        if (!mBit) { // After 128 was shifted by 
            if (mBytePos >= mBufSize ) {
                throw CEndOfMem();
            }
            mpBuf[mBytePos++] = mChar;
            mBit = 1;
            mChar = 0;
        }
        if (v) mChar |= mBit;
        mBit <<= 1;    						
    }
    inline void PutByte(unsigned char c) {
        if (mBytePos >= mBufSize ) {
            throw CEndOfMem();
        }
        mpBuf[mBytePos++] = c;
    }
    inline void FinishOutput() {
        if (mBit!=1) {
            mpBuf[mBytePos++]= mChar;
        }
    }
    inline size_t getPos() {return mBytePos;}
};

class CBitReader {
private:
    size_t         mBufSize;
    size_t         mReadLen;
    size_t         mBytePos;
    char *         mpBuf;
    unsigned char  mBit;
    unsigned char  mChar;
    int            mFd;
    bool           mEof;
    // throws CReachedEof on EOF
    inline void readBuf() {
        mBytePos = 0;
        mBit	= 1;	
        mReadLen = read (mFd, mpBuf, mBufSize);
        if (mReadLen < 0) {
            THROW_RUNTIME_ERROR(SysError());
        }
        if (!mReadLen) {
            mEof = true;
            throw CReachedEof();
        }
        mChar = mpBuf[mBytePos]; 
    }
public:
    CBitReader(size_t bufSize) : mBufSize(bufSize), mpBuf(new char[bufSize]), mReadLen(0), mFd(-1), mEof(true){}
    ~CBitReader(){delete [] mpBuf;}
    bool isEof(){return mEof;}
    inline void StartInput(int fd) {
        mFd = fd;
        mEof = false;
        readBuf();
    }
    // throws CReachedEof on EOF
    inline void GetBit(bool & v) {
        if (mBit> 8) {
            mBytePos++;
            if (mBytePos < mReadLen ) {
                mBit  = 1;
                mChar = mpBuf[mBytePos];
            } else {
                readBuf();
            }
        } 
        v = mChar & 0x1;
        mChar >>= 1;
        mBit++;
    }
    // throws CReachedEof on EOF
    inline unsigned char GetByte() {
        if (mBytePos >= mReadLen) {
            readBuf();
        }
        return mpBuf[mBytePos++];
    }
};

class CMemBitReader {
private:
    size_t         mBufSize;
    size_t         mBytePos;
    const char *   mpBuf;
    unsigned char  mBit;
    unsigned char  mChar;
    inline void StartInput() {
        mBytePos = 0;
        mBit	= 1;	
        mChar = mpBuf[mBytePos]; 
    }
public:
    CMemBitReader(const char *pBuf, size_t bufSize) : mBufSize(bufSize), mpBuf(pBuf) {
        if (!bufSize) {
            THROW_RUNTIME_ERROR("Zero buffer is not allowed");
        }
        StartInput();
    }
    inline void GetBit(bool & v) {
        if (mBit> 8) {
            mBytePos++;
            if (mBytePos < mBufSize ) {
                mBit  = 1;
                mChar = mpBuf[mBytePos];
            } else {
                throw CReachedEof();
            }
        } 
        v = mChar & 0x1;
        mChar >>= 1;
        mBit++;
    }
    inline unsigned char GetByte() {
        if (mBytePos >= mBufSize) {
            throw CReachedEof();
        }
        return mpBuf[mBytePos++];
    }
};

/* 
 * Bellow goes Elias Gamma coding/decoding
 * Note that GammaEncode codes only non-zero integers!
 * Name is the name of the output buffer
 * GammaEncode after writing [log_2(v)] 
 * zeroes following by 1 
 * writes to the bitstream digits of v in reveresed order
 * less the most significant digit,
 * that is starting from the the least significant bit
 */
template <class T, class BitWriter>
inline void GammaEncode(T v, BitWriter & writer) {
    register T     vn   = v>>1;
    register int   logv = 0;
    for(; vn; logv++, vn>>=1) {
        writer.PutBit(0);
    }
    writer.PutBit(1);
    vn = (v);

    for(register int i = logv; i; i--, vn>>=1) {
        writer.PutBit(vn & 0x1);
    }
}

template <class T, class BitReader >
inline void GammaDecode(T & v, BitReader & reader) {
    register bool      b = false;
    register uint32    n = 0;

    for(;!b;) { 
        reader.GetBit(b);
        n += 1 - int(b);
    }
    register uint32 vn = 1;
    for (v = 0; n; n--, vn <<= 1) {
        reader.GetBit(b);
        if (b) v |= vn;
    }
    v |= vn; // The most significant bit is not explicitly coded!
}

/* 
 * Below goes the Elias Delta Coding/Decoding
 * Note that DeltaEncodeIncl codes zero integer 
 * (as well as positive ones).
 * First codes [log_2(v)]+1 using GammaEncode
 * Then stores digits of v in reversed order
 * less the most significant digit starting from the
 * least significant bit
 */
template <class T, class BitWriter>
inline void DeltaEncode(T v, BitWriter & writer) {
    register T     vn_   = v>>1;
    register int   logv_ = 1;
    for(;vn_ ; logv_++, vn_>>=1){
    }
    /* Got [log_2(v)]+1 !             
     * Now store it using GammaEncode
     */
    GammaEncode(logv_, writer);
    // Now do as with GammaEncode method
    vn_ = v;
    for(register int i = logv_ - 1;i ;i-- , vn_>>=1) {
        writer.PutBit(vn_ & 0x1);
    }
}

template <class T, class BitReader>
inline void DeltaDecode(T & v, BitReader & reader) {
    register bool      b_= 0;
    register uint32    n_ = 0;
    register uint32    vn_ = 1;

    GammaDecode(n_, reader);
    n_--;
    for(v = 0; n_; n_--, vn_<<=1) {
        reader.GetBit(b_);
        if (b_) v |= vn_;
    }
    v |= vn_; // The most significant bit is not implicitly coded!
}
/*
 * Below go variable byte encoding routing.
 * It is worth noting, that they are much 
 * faster
 */
#include <iostream>
template <class T, class BitWriter>
inline void VarByteEncode(T v, BitWriter & writer) {
    for (register int t = v; t;) {
        register unsigned char b = (t & 127);

        t >>= 7;
        if (t) {
            b |= 128;
        }
        writer.PutByte(b);
    }
}

template <class T, class BitReader>
inline void VarByteDecode(T & v, BitReader & reader) {
    register unsigned char c;
    register unsigned int shift = 0;
    for (v = 0;;shift += 7) {
        c = reader.GetByte();

        v += ((c & 127) << shift);

        if (!(c & 128)) {
            break;
        }
    }
}

#endif
