#include "index.h"
#include "langcase.h"
#include "myexcept.h"

CFreqKdTreeIndex::CFreqKdTreeIndex (bool hasTransp, EAlphabet AlphaType)
: CInterface(hasTransp, AlphaType), MaxFuzzy(GetMaxFuzzy()), opened (false)
{
    const char *pABC = GetAlphabet(AlphaType);

    if (!pABC) {
        THROW_AND_PRINT_RUNTIME_ERROR("Undefined alphabet for code: " << AlphaType);
    }

    int charNum = strlen(pABC);

    char letSmall[charNum];
    char letCaps[charNum];

    for (unsigned int i = 0; i < charNum; ++i) {
        letSmall[i] = tolower_eng_rus(pABC[i]);
        letCaps[i] = toupper_eng_rus(pABC[i]);
    }

    bool printAddStat = GetEnvSwitch("PRINT_ADD_STAT");

    int VectorLength = MaxFuzzy;
    
    for (unsigned int i = 0; i < charNum; ++i) {
        int pos = i % VectorLength;

        if (i < VectorLength) {
            vMap.AddChar(letSmall[i]);
            vMap.AddCharAlias(letSmall[i], letCaps[i]);
        } else {
            vMap.AddCharAlias(letSmall[pos], letSmall[i]);
            vMap.AddCharAlias(letSmall[pos], letCaps[i]);
        }
    }

    kdAlloc = new CFastAllocator(sizeof *kd);
    vecAlloc = new CFastAllocator(MaxFuzzy * sizeof(VectorType));
    void * p = kdAlloc->allocChunk();
    kd = new (p) kdtree (*kdAlloc, *vecAlloc, MaxFuzzy, NULL, NULL);
}

int CFreqKdTreeIndex::GetMaxFuzzy()
{
    int res = _VECTOR_LENGTH;
    GetEnv("VECTOR_LENGTH", res);
    
    if (GetEnvSwitch("PRINT_ADD_STAT")) {
        cout << "Vector size: " << res << endl;
    }
    return res;
}

void 
CFreqKdTreeIndex::StringToVec(const char *str, unsigned int vecSize, VectorType * vec)
{
    unsigned int val = 0;
    int len = strlen (str);

    memset (vec, 0, vecSize * sizeof (*vec));

    if (vecSize == 0)
        return;

    for (const char *ps = str; *ps; ps++) {
        unsigned char c = (unsigned char) *ps;
        int num = vMap.CharNum(c);
        vec[num]++;
    }
}

CFreqKdTreeIndex::~CFreqKdTreeIndex ()
{
    if (opened)
        CloseDB ();
    //delete kd;
    kd->~kdtree();
    delete kdAlloc;
    delete vecAlloc;
}

bool CFreqKdTreeIndex::Insert (const char *str)
{
    VectorType    vec[MaxFuzzy];

    dict.push_back(string(str));

    StringToVec(str, MaxFuzzy, vec);
    kd->insertNodeIntoTree (vec, dict.back().c_str(), NULL);

    return true;
}

bool CFreqKdTreeIndex::Search (const char *str, int kerr, int *numComp,
                           StringList * resList, bool silent)
{
    myagrep        a(str, kerr, mHasTransp);
    VectorType     vec[MaxFuzzy];
    StringList     res;

    unsigned int   dist = kerr;

    StringList *res1 = (resList == NULL) ? &res : resList;

    StringToVec(str, MaxFuzzy, vec);

    kd->Search (vec, dist, a, res1, numComp);
    if (!silent) {
        for (StringList::iterator i = res1->begin (); i != res1->end ();
             i++) {
            cout << *i << endl;
        }
    }
    return true;
}

void CFreqKdTreeIndex::MakeBalanced ()
{
    kdtree * oldTree = kd;

    //std::cerr << "Data size: " << kd->dataSize() << std::endl;;
    CFastAllocator * newKdAlloc = new CFastAllocator(sizeof *kd);
    CFastAllocator * newVecAlloc = new CFastAllocator(MaxFuzzy * sizeof(VectorType));

#if 0
    std::cout << "Before balanced, consumed: " << kdAlloc->getAllocatedMem() << ", " << vecAlloc->getAllocatedMem() << std::endl;
#endif
    kd = oldTree->getBalancedCopy(*newKdAlloc, *newVecAlloc);
#if 0
    std::cerr << "After balanced, new consumed: " << newKdAlloc->getAllocatedMem() << ", " << newVecAlloc->getAllocatedMem() << std::endl;
#endif
#if 0
    oldTree->printTree();
    cout << " ==================== \n";
    kd->printTree();
#endif

    //delete oldTree;
    oldTree->~kdtree();
    delete kdAlloc;
    delete vecAlloc;

    kdAlloc = newKdAlloc;
    vecAlloc = newVecAlloc;
}

bool CFreqKdTreeIndex::SearchInDB (const char *str, int kerr, int *numComp,
                               StringList * resList, bool silent)
{
    unsigned int dist = kerr;

    if (!opened) {
        cerr << "Database is not opened!" << endl;
    }
    myagrep       a(str, kerr, mHasTransp);
    VectorType    vec[MaxFuzzy];
    StringList    res;
    StringList *  res1 = (resList == NULL) ? &res : resList;

    StringToVec(str, MaxFuzzy, vec);
    //cout<<str<<" :";PrintNgramVec (MaxFuzzy, vec);

    if (!kdtree::SearchInFile (fd, vec, dist, a, res1, numComp)) {
        return false;
    }
    if (!silent) {
        for (StringList::iterator i = res1->begin (); i != res1->end ();
             i++) {
            cout << *i << endl;
        }
    }
    return true;
}

bool CFreqKdTreeIndex::OpenDB (const char *file_name, bool create)
{
    if (opened)
        CloseDB ();
    if (!create) {
        fd = open (file_name, O_RDONLY, 0);
    } else {
        fd = open (file_name, O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
    }
    if (fd < 0) {
        Perror ("Error opening file");
        return false;
    }
    return (opened = true);
}

bool CFreqKdTreeIndex::CloseDB ()
{
    if (!opened)
        return true;
    if (close (fd) != 0)
        return false;
    opened = false;
    return true;
}

bool CFreqKdTreeIndex::DumpDB (const char *file_name)
{
    bool stat;
    if (opened)
        CloseDB ();
    if (!OpenDB (file_name, true)) {
        return false;
    }
    stat = kd->DumpToFile (fd, 0);
    return stat && CloseDB ();
}

/*
** This function substitutes the function from a base class.
*/
bool CFreqKdTreeIndex::CreateInMemIndex (const char *inputFile,
                                       int insertCount, int *dcnt,
                                       bool silent)
{
    *dcnt = 0;
    string s;

    ifstream f (inputFile);
    StringList lst;
    int lstCnt = 0;
    while (f >> s) {
        toUpper (s);
        if (insertCount == 1) {
            Insert (s.c_str ());
        } else {
            lst.push_back (s);
            lstCnt++;
            if (lstCnt >= insertCount) {
                this->InsertList (lst);
                lstCnt = 0;
                lst.clear ();
            }
        }
        if (dcnt)
            (*dcnt)++;
        if (!silent && !((*dcnt) % 10000))
            cout << (*dcnt) << "\n";
    }
    if (lstCnt)
        this->InsertList (lst);
    MakeBalanced ();

    return true;
}
