#include "index.h"
#include "myexcept.h"

CIndex::CIndex(bool hasTransp, EAlphabet AlphaType) : mHasTransp(hasTransp), mAlphaType(AlphaType), 
                        MaxFuzzy(GetMaxFuzzy())
{
    const char *pABC = GetAlphabet(AlphaType);

    if (!pABC) {
        THROW_AND_PRINT_RUNTIME_ERROR("Undefined alphabet for code: " << AlphaType);
    }

    int charNum = strlen(pABC);

    char letSmall[charNum];
    char letCaps[charNum];

    for (unsigned int i = 0; i < charNum; ++i) {
        letSmall[i] = tolower_eng_rus(pABC[i]);
        letCaps[i] = toupper_eng_rus(pABC[i]);
    }

    int VectorLength = MaxFuzzy;
    
    for (unsigned int i = 0; i < charNum; ++i) {
        int pos = i % VectorLength;

        if (i < VectorLength) {
            vMap.AddChar(letSmall[i]);
            vMap.AddCharAlias(letSmall[i], letCaps[i]);
        } else {
            vMap.AddCharAlias(letSmall[pos], letSmall[i]);
            vMap.AddCharAlias(letSmall[pos], letCaps[i]);
        }
    }

    kdAlloc = new CFastAllocator(sizeof *kd);
    vecAlloc = new CFastAllocator(MaxFuzzy * sizeof(VectorType));
    void * p = kdAlloc->allocChunk();
    kd = new (p) kdtree (*kdAlloc, *vecAlloc, MaxFuzzy, NULL, NULL);
}

void CIndex::MakeBalanced()
{
    kdtree * oldTree = kd;

    CFastAllocator * newKdAlloc = new CFastAllocator(sizeof *kd);
    CFastAllocator * newVecAlloc = new CFastAllocator(MaxFuzzy * sizeof(VectorType));

#if 0
    std::cout << "Before balanced, consumed: " << kdAlloc->getAllocatedMem() << ", " << vecAlloc->getAllocatedMem() << std::endl;
#endif
    kd = oldTree->getBalancedCopy(*newKdAlloc, *newVecAlloc);
#if 0
    std::cerr << "After balanced, new consumed: " << newKdAlloc->getAllocatedMem() << ", " << newVecAlloc->getAllocatedMem() << std::endl;
#endif
#if 0
    oldTree->printTree();
    cout << " ==================== \n";
    kd->printTree();
#endif

    //delete oldTree;
    oldTree->~kdtree();
    delete kdAlloc;
    delete vecAlloc;

    kdAlloc = newKdAlloc;
    vecAlloc = newVecAlloc;
}

void CIndex::Traverse(const string &str, int d, vd &stat)
{
    VectorType    vec[MaxFuzzy];
    vi            cnt, prob;


    StringToVec(str.c_str(), MaxFuzzy, vec);
    kd->Traverse(vec, d, prob, cnt);

    assert(prob.size() == cnt.size());

    int N = prob.size();

    stat.resize(N);

    for (int i = 0; i < N; ++i) {
        stat[i] = double(prob[i]) / cnt[i];
    }
}

void CIndex::Insert(const string &str)
{
    VectorType    vec[MaxFuzzy];


    StringToVec(str.c_str(), MaxFuzzy, vec);
    kd->insertNodeIntoTree(vec, str.c_str(), NULL);
}

bool CIndex::CreateIndex (const char *pInputFile)
{
    string s;

    ifstream f (pInputFile);

    if (!f) {
        return false;
    }
    StringList lst;

    while (f >> s) {
        toUpper(s);

        Insert(s);
    }
    MakeBalanced();
    
    return true;
}

CIndex::~CIndex ()
{
    kd->~kdtree();
    delete kdAlloc;
    delete vecAlloc;
}

void 
CIndex::StringToVec(const string &str, unsigned int vecSize, VectorType * vec)
{
    unsigned int val = 0;

    memset (vec, 0, vecSize * sizeof (*vec));

    if (vecSize == 0)
        return;

    for (string::const_iterator ps = str.begin(); ps != str.end(); ++ps) {
        unsigned char c = (unsigned char) *ps;
        int num = vMap.CharNum(c);
        vec[num]++;
    }
}

int CIndex::GetMaxFuzzy()
{
    int res = _VECTOR_LENGTH;
    GetEnv("VECTOR_LENGTH", res);
    
    if (GetEnvSwitch("PRINT_ADD_STAT")) {
        cout << "Vector size: " << res << endl;
    }
    return res;
}
