#ifndef _LEN1INDEX_H
#define _LEN1INDEX_H

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <string>
#include <iostream>
#include <fcntl.h>
#include <errno.h>
#include <fstream>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>

#include <vector>
#include <map>
#include <set>
#include <algorithm>

#include "interface.h"
#include "myagrep.h"
#include "ngram.h"


typedef vector < const char *>     PtrListVector;

// TODO/TBD currently only array lists are supported

#ifdef _HASH_LIST
    #include <hashport>
    typedef hash_set < const char *>   PtrList;
    typedef hash_set < off_t >         OffsetList;
#else
    #ifdef _TREE_LIST
        typedef set < const char *>        PtrList;
        typedef set < off_t >              OffsetList;
    #else
        typedef vector < const char *>     PtrList;
        typedef vector < off_t >           OffsetList;
    #endif
#endif

typedef PtrList::iterator PtrListIterator;
typedef OffsetList::iterator OffsetListIterator;

struct EntryDataCompressInMem {
    const char *   mpStart;
    size_t         cnt;
    size_t         size;
    EntryDataCompressInMem(const char *pStart = NULL, size_t cnt = 0, size_t size= 0) :
            mpStart(pStart), cnt(cnt), size(size) {}
};

typedef map < u_int64_t, PtrList * >  mPtrList;
typedef map < u_int64_t, EntryDataCompressInMem >  mPtrListComp;
typedef map < const char *, off_t >   mOffset;


// ListInsert keeps list in a sorted order
template <class T> void ListInsert (vector <T> *, const T elem);

// These 2 functions implement intersection and union
template <class T>
    void ListIntersection (vector <T> &first, vector <T> &second,
                           vector <T> *result);
template <class T> void ListUnion (vector <T> &first,
                                     vector <T> &second,
                                     vector <T> *result);

template <class T>
    void CopyList (vector <T> *dst, const vector <T> &src);
#ifdef _DEBUG
template <class T> void PrintList (const T & p);
#endif

class CLen1Index {
  public:
    static const int MaxIntersectCount = 2;
    static const int pagesize =  2 * 1024;
    static const int cachesize = 0;//32 * 1024;

    static unsigned int SeqIndexBound(int kerr) {
        return (1 + kerr);
    }
    void setUseDelta(bool useDelta) { mUseDelta = useDelta;}
    virtual ~CLen1Index();
  protected:
    mPtrList              invgram;
    mPtrListComp          invgramComp;
    int                   NgramLen;
    int                   fd;
    bool                  opened;
    off_t *               lstStart;
    off_t                 mapSize;
    static unsigned int   fetchTime;
    static int            recFetched;
    static int            lstFetched;
    bool                  mHasTransp;
    bool                  mHasCompress;
    bool                  mIsCompInMem;
    bool                  mUseDelta; // use delta compression
    double                mAvgGapSize;
    double                mAvgListSize;

    virtual bool ReadListComp(const char *str, int StrLen, int minpos, int maxpos, OffsetList * ptr);
  public:
    CLen1Index(int _NgramLen, bool hasTransp, bool hasCompress, bool useDelta):
            NgramLen (_NgramLen), opened (false), mHasTransp(hasTransp), 
            mHasCompress(hasCompress), mIsCompInMem(false), 
            mUseDelta(useDelta),
            mapSize(0), lstStart(NULL), mAvgGapSize(0.0), mAvgListSize(0.0) {}
    virtual void ReadList (const char *str, int StrLen, int minpos, int maxpos, PtrList * ptr);
    virtual int Insert (const char *str);
    virtual int Search (const char *str, int kerr, myagrep & a, int *nComp,
                StringList * resList, bool silent = false);

    virtual bool OpenDB (const char *name, bool create = false) = 0;
    virtual bool CloseDB () = 0;
    virtual bool DumpDB (const char *index_name, mOffset &) = 0;
    virtual bool CompressInMem (mOffset &);

    virtual bool ReadListDB (const char *str, int StrLen, int minpos, int maxpos, OffsetList * ptr) = 0;
    virtual bool SearchInDB(const char *dict, const char *str, const myagrep & a, int kerr, int *nComp, StringList * resList, bool silent = false);
    virtual bool SearchComp(const char *dict, const char *str, const myagrep & a, int kerr, int *nComp, StringList * resList, bool silent = false);

    static void DisplayStat () {
        cout << " Lists fetched: " << lstFetched << " Records fetched: " <<
            recFetched << " Time fetched: " << fetchTime << endl;
        if (lstFetched > 0) {
            cout << " Average records fetched per list:" <<
                double (recFetched) / lstFetched << endl;
            cout << " Average fetch time per list:" << double (fetchTime) /
                lstFetched << endl;
        }
    }
    virtual void ResetCounters(){
         RESET_COUNTER(check_string);
    }
    virtual void PrintAddStat(int testNum, int kerr, ostream &) = 0;
    DECLARE_STATIC_COUNTER(check_string);
#ifdef _DEBUG
    virtual void Print ();
#endif
};

template <class T> 
struct DataAndPos {
    T   mData;
    int mPos;
    DataAndPos(T data = T(), int pos = 0) : mData(data), mPos(pos) {
    }
    bool operator <(const DataAndPos & o) const {
      if (mData != o.mData) return mData < o.mData;
      return mPos < o.mPos;
    }
};

#ifdef _HASH_LIST

template < class T > void ListInsert (hash_set < T > *p, const T elem)
{
    p->insert (elem);
}

template < class T >
    void ListUnion (hash_set < T > &first, hash_set < T > &second,
                    hash_set < T > *result)
{
    result->clear ();
    hash_set < T >::iterator i;
    for (i = first.begin (); i != first.end (); i++)
        result->insert (*i);
    for (i = second.begin (); i != second.end (); i++)
        result->insert (*i);
}

template < class T >
    void ListIntersection (hash_set < T > &first, hash_set < T > &second,
                           hash_set < T > *result)
{
    result->clear ();
    for (hash_set < T >::iterator i = first.begin (); i != first.end ();
         i++)
        if (second.find (*i) != second.end ())
            result->insert (*i);
}

#else
#ifdef _TREE_LIST
template < class T > void ListInsert (set < T > *p, const T elem)
{
    p->insert (elem);
}

template < class T >
    void ListUnion (set < T > &first, set < T > &second, set < T > *result)
{
    result->clear ();
    typename set < T >::iterator i;
    for (i = first.begin (); i != first.end (); i++)
        result->insert (*i);
    for (i = second.begin (); i != second.end (); i++)
        result->insert (*i);
}

template < class T >
    void ListIntersection (set < T > &first, set < T > &second,
                           set < T > *result)
{
    result->clear ();
    for (typename set < T >::iterator i = first.begin (); i != first.end (); i++)
        if (second.find (*i) != second.end ())
            result->insert (*i);
}
#else
template < class T > void ListInsert (vector < T > *p, const T elem)
{
    // The list is kept in ascending order
#if 0
    typename vector < T >::iterator it = p->begin ();
    for (; it != p->end () && (*it) < elem; ++it);
#else
    typename vector < T >::iterator it = p->end();

    if (!p->empty() && elem <= p->back()) {
        it = lower_bound(p->begin (), p->end(), elem);
    }

    if (it != p->end() && *it == elem) {
        // Duplicates are ignored
        return;
    }

    if (p->capacity() <= p->size() + 1) {
        p->reserve(p->size() + p->size() >> 3 + 1);
    }

    p->insert (it, elem);
#endif
}

template < class T >
    void ListIntersection (vector < T > &first, vector < T > &second,
                           vector < T > *result)
{
    result->clear ();
    typename vector < T >::iterator
        first1 = first.begin (),
        first2 = second.begin (), last1 = first.end (), last2 =
        second.end ();

    while (first1 != last1 && first2 != last2)
        if (*first1 < *first2)
            ++first1;
        else if (*first2 < *first1)
            ++first2;
        else {
            result->insert (result->end (), *first1);
            ++first1;
            ++first2;
        }
}


template < class T >
    void ListUnion (vector < T > &first, vector < T > &second,
                    vector < T > *result)
{
    result->clear ();
    typename vector < T >::iterator
        first1 = first.begin (),
        first2 = second.begin (), last1 = first.end (), last2 =
        second.end ();

    while (first1 != last1 && first2 != last2) {
        if (*first1 < *first2) {
            result->insert (result->end (), *first1);
            ++first1;
        } else if (*first2 < *first1) {
            result->insert (result->end (), *first2);
            ++first2;
        } else {
            result->insert (result->end (), *first1);
            ++first1;
            ++first2;
        }
    }
    while (first1 != last1) {
        result->insert (result->end (), *first1++);
    }
    while (first2 != last2) {
        result->insert (result->end (), *first2++);
    }
}
#endif
#endif

#ifdef _DEBUG
template < class T > void PrintList (const T & p)
{
    PtrList::const_iterator i;
    for (i = p.begin (); i != p.end (); i++)
        cout << "#" << int (*i) << " " << *i << "\n";
    cout << "*************\n";
}
#endif

#endif
