#include "len1indexBDB.h"
#include <utility>
#include <sys/timeb.h>
#include <db_init.h>
#include "format.h"
#include "int_coder_inl.h"

const char * extDict = NULL;

DEFINE_CLASS_STATIC_COUNTER(CLen1IndexBDB, load_list)

bool CLen1IndexBDB::OpenDB(const char *name, bool create)
{
    if (opened)
        CloseDB ();

    if (!dbenv) {
         dbenv = db_init (NULL, CLen1Index::cachesize);
    }
    if (!dbenv) {
        cerr << "Cannot db_init" << endl;
         return false;
    }

    if ((errno = db_create (&dbp, dbenv, 0)) != 0) {
        cerr << "Error creating database : " << db_strerror (errno) << endl;
        return false;
    }

    if ((errno = dbp->set_pagesize (dbp, CLen1Index::pagesize)) != 0) {
        cerr << "Error setting page size : " << db_strerror (errno) << endl;
        return false;
    }

    /* Opening the database. */
    if ((errno = dbp->open (dbp, NULL, name, NULL, DB_BTREE,
                            ((!create) ? DB_RDONLY : DB_CREATE |
                             DB_TRUNCATE), CInterface::defaultAccess))) {
        cerr << "Error opening database '" << name << "' create?: " << create<< " error msg:" << db_strerror (errno) << endl;
        return false;
    }
    /* Opening file */
    string fname = string (name) + ".lst";
    if ((fd = open (fname.c_str (),
                    ((!create) ? O_RDONLY : O_CREAT | O_TRUNC | O_RDWR),
                    CInterface::defaultAccess)) <= 0) {
        Perror (Format ("Cannot open file %s", fname.c_str ()));
        return false;
    }
    lstStart = NULL;
    mapSize = 0;
    if (!create) {
        mapSize = FileSize (fname.c_str ());

        if (mapSize < 0) {
            Perror ("Cannot determine file size");
        }
        mapSize = CInterface::getPageAlignedValue(mapSize);

        if (mapSize) {
            lstStart = (off_t *) mmap (NULL, mapSize, PROT_READ, MAP_SHARED, fd, 0);
            if (lstStart == MAP_FAILED) {
                Perror ("Map failed" + string(" File name: '") + fname + "'");
                return false;
            }
        } else {
            lstStart = NULL;
        }
    }

    return opened = true;
}

bool CLen1IndexBDB::CloseDB ()
{
    if (!opened)
        return true;
    if ((errno = dbp->close (dbp, 0)) != 0) {
        cerr << db_strerror (errno);
        return opened = false;
    }
    if (mapSize != 0 && lstStart) {
        if (munmap (lstStart, mapSize)) {
            Perror ("Cannot unmap");
        }
    }
    if (close (fd)) {
        Perror ("Cannot close file");
        return opened = false;
    }
    db_close(dbenv);
    dbenv = NULL;
    opened = false;
    fd = 0;
    return true;
}


bool CLen1IndexBDB::ReadListDBNoCompress(const char *str, int StrLen, int minpos,
                             int maxpos, OffsetList * ptr)
{
    if (!lstStart) {
        return true; // Empty index
    }
    if (minpos < 0) minpos = 0;
    bool first = true;
    OffsetList pl[2];
    int nc = 0;
    timeb t1, t2;

    assert (str);
    ptr->clear ();

    START_COUNTER(load_list)

    lstFetched++;
    if (NgramLen > StrLen) {
        // create a union of all lists
        // We will iterate through the map using minimum and maximum ngram value
        // and minimum/maximum values of position/length
        char minstr[NgramLen], maxstr[NgramLen];
        char mins[NgramLen + 2], maxs[NgramLen + 2];
        int i = 0;
        // filling common part
        for (; i < StrLen; i++)
            minstr[i] = maxstr[i] = str[i];
        for (; i < NgramLen; i++)
            minstr[i] = 32, maxstr[i] = 255;

        ngram gmin, gmax;
        gmin.pack (minstr, NgramLen, minpos);
        gmax.pack (maxstr, NgramLen, maxpos);
        gmin.toStr (mins, NgramLen);
        gmax.toStr (maxs, NgramLen);
        DBC *dbcp;
        DBT key, data;
        memset (&key, 0, sizeof (key));
        memset (&data, 0, sizeof (data));
        key.data = mins;
        key.size = NgramLen + 2;
        ftime (&t1);
        if ((errno = dbp->cursor (dbp, NULL, &dbcp, 0))) {
            cerr << strerror (errno) << "\n";
            STOP_COUNTER(load_list)
            return false;
        }
        unsigned int flags = DB_SET_RANGE;
        OffsetList p;
        for (;;) {
            errno = dbcp->c_get (dbcp, &key, &data, flags);
            flags = DB_NEXT;
            if (errno) {
                if (errno == DB_NOTFOUND)
                    break;
                else {
                    cerr << strerror (errno) << "\n";
                    STOP_COUNTER(load_list)
                    return false;
                }
            }
            // If greater than the last possible ngram than break
            if (strncmp ((char *) key.data, maxs, key.size) > 0)
                break;
            // Now check whether pos is ok
            int pos = ngram::PosFromStr ((char *) key.data, NgramLen);

            if (pos <= maxpos && pos >= minpos) {
                size_t cnt = *((off_t *) data.data);
                size_t listIndex = *(1 + (off_t *) data.data);

                // Getting in-list offsets
                off_t *ptr = lstStart + listIndex;
				unsigned int sz = p.size ();
				p.resize (sz + cnt);
                for (int j = 0; j < cnt; j++, ptr++) {
					p[sz + j] = *ptr;
                    recFetched++;
                }
            }
        }
		sort (p.begin (), p.end ());
        ftime (&t2);
        fetchTime +=
            1000 * (t2.time - t1.time) + (t2.millitm - t1.millitm);
        if ((errno = dbcp->c_close (dbcp))) {
            cerr << strerror (errno) << "\n";
            STOP_COUNTER(load_list)
            return false;
        }
        // Joining lists
        *ptr = p;
        STOP_COUNTER(load_list)
        return true;
    } else {
        typedef  DataAndPos<off_t>   PPOS;
        typedef  vector<PPOS>        offList;

        offList                      tmpList[2];

        nc = 0;
        
        // Create an intersection of all lists (< MaxListCount)
        for (int i = 0; i <= min (CLen1Index::MaxIntersectCount, StrLen - NgramLen); i++) {
            char   mins[NgramLen + 2], maxs[NgramLen + 2];
            ngram  gmin, gmax;

            gmin.pack (&str[i], NgramLen, minpos + i);
            gmax.pack (&str[i], NgramLen, maxpos + i);

            gmin.toStr (mins, NgramLen);
            gmax.toStr (maxs, NgramLen);

            DBC *dbcp;
            DBT key, data;
            memset (&key, 0, sizeof (key));
            memset (&data, 0, sizeof (data));
            key.data = mins;
            key.size = NgramLen + 2;
            ftime (&t1);

            if ((errno = dbp->cursor (dbp, NULL, &dbcp, 0))) {
                cerr << strerror (errno) << "\n";
                STOP_COUNTER(load_list)
                return false;
            }

            unsigned int flags = DB_SET_RANGE;

            offList p;

            for (;;) {
                errno = dbcp->c_get (dbcp, &key, &data, flags);
                flags = DB_NEXT;
                if (errno) {
                    if (errno == DB_NOTFOUND)
                        break;
                    else {
                        cerr << strerror (errno) << "\n";
                        STOP_COUNTER(load_list)
                        return false;
                    }
                }
                // If greater than the last possible ngram than break
                if (strncmp ((char *) key.data, maxs, key.size) > 0)
                    break;
                // Now check whether pos is ok
                // Do not forget that we are incrementing acceptable pos value by variable i
                int pos = ngram::PosFromStr ((char *) key.data, NgramLen);
                if (pos <= maxpos + i && pos >= minpos + i) {
                    char tmp[NgramLen + 1];
                    strncpy(tmp, (char *) key.data, NgramLen);
                    tmp[NgramLen] = 0;
#if 0
                    cout << "Pos of '" << tmp << "'" << pos << " of words: " ;
#endif

                    size_t cnt = *((off_t *) data.data);
                    size_t listIndex = *(1 + (off_t *) data.data);

                    off_t *ptr = lstStart + listIndex;
					unsigned int sz = p.size ();
					//p.resize (sz + cnt);
					p.reserve(sz + cnt);

                    for (int j = 0; j < cnt; j++, ptr++) {
						//p[sz + j] = *ptr;
                        p.push_back(PPOS(*ptr, pos - i));
#if 0
                        cout << &extDict[*ptr] << ", ";
#endif
                        recFetched++;
                    }
#if 0
                    cout << endl;
#endif
                }
            }
			sort (p.begin (), p.end ());
            ftime (&t2);
            fetchTime +=
                1000 * (t2.time - t1.time) + (t2.millitm - t1.millitm);
            if ((errno = dbcp->c_close (dbcp))) {
                cerr << strerror (errno) << "\n";
                STOP_COUNTER(load_list)
                return false;
            }
            if (first) {
                first = false;
                tmpList[1 - nc] = p;
                nc = 1 - nc;
            } else {
                ListIntersection (tmpList[nc], p, &tmpList[1 - nc]);
                nc = 1 - nc;
            }
            if (tmpList[nc].size() <= 1) {
                break;
            }
        }
        //*ptr = pl[nc];
        ptr->reserve(tmpList[nc].size());
        for (unsigned int k = 0; k < tmpList[nc].size(); ++k)
            ptr->push_back(tmpList[nc][k].mData);
        STOP_COUNTER(load_list)
        return true;

    }
    cerr << "We mustn't reach this point (internal error)\n";
    STOP_COUNTER(load_list)
    return false;

}

bool CLen1IndexBDB::ReadListDBCompress(const char *str, int StrLen, int minpos,
                             int maxpos, OffsetList * ptr)
{
    if (!lstStart) {
        return true; // Empty index
    }
    if (minpos < 0) minpos = 0;
    bool first = true;
    OffsetList pl[2];
    int nc = 0;
    timeb t1, t2;

    assert (str);
    ptr->clear ();

    START_COUNTER(load_list)

    lstFetched++;
    if (NgramLen > StrLen) {
        // create a union of all lists
        // We will iterate through the map using minimum and maximum ngram value
        // and minimum/maximum values of position/length
        char minstr[NgramLen], maxstr[NgramLen];
        char mins[NgramLen + 2], maxs[NgramLen + 2];
        int i = 0;
        // filling common part
        for (; i < StrLen; i++)
            minstr[i] = maxstr[i] = str[i];
        for (; i < NgramLen; i++)
            minstr[i] = 32, maxstr[i] = 255;

        ngram gmin, gmax;
        gmin.pack (minstr, NgramLen, minpos);
        gmax.pack (maxstr, NgramLen, maxpos);
        gmin.toStr (mins, NgramLen);
        gmax.toStr (maxs, NgramLen);
        DBC *dbcp;
        DBT key, data;
        memset (&key, 0, sizeof (key));
        memset (&data, 0, sizeof (data));
        key.data = mins;
        key.size = NgramLen + 2;
        ftime (&t1);
        if ((errno = dbp->cursor (dbp, NULL, &dbcp, 0))) {
            cerr << strerror (errno) << "\n";
            STOP_COUNTER(load_list)
            return false;
        }
        unsigned int flags = DB_SET_RANGE;
        OffsetList p;
        for (;;) {
            errno = dbcp->c_get (dbcp, &key, &data, flags);
            flags = DB_NEXT;
            if (errno) {
                if (errno == DB_NOTFOUND)
                    break;
                else {
                    cerr << strerror (errno) << "\n";
                    STOP_COUNTER(load_list)
                    return false;
                }
            }
            // If greater than the last possible ngram than break
            if (strncmp ((char *) key.data, maxs, key.size) > 0)
                break;
            // Now check whether pos is ok
            int pos = ngram::PosFromStr ((char *) key.data, NgramLen);

            if (pos <= maxpos && pos >= minpos) {
                size_t listSize = *((off_t *) data.data);
                size_t listOff = *(1 + (off_t *) data.data);
                size_t listCnt = *(2 + (off_t *) data.data);

                // Getting in-list offsets
                char *ptr = reinterpret_cast<char *>(lstStart); 
                ptr += listOff;

                CMemBitReader      reader(ptr, listSize);

                // Getting compressed in-list offsets
                try {

		            p.reserve(p.size() + listCnt);

                    off_t last = 0;
                    off_t t;
                    if (!mUseDelta) {
                        for (int j = 0; j < listCnt; j++) {
                            VarByteDecode(t, reader);
                            t--;
                            last += t;
		                    p.push_back(last);
                            recFetched++;
                        }
                    } else {
                        for (int j = 0; j < listCnt; j++) {
                            DeltaDecode(t, reader);
                            t--;
                            last += t;
		                    p.push_back(last);
                            recFetched++;
                        }
                    }
                } catch (const CReachedEof &) {
                    cerr << "unexpected EOF" << endl;
                    STOP_COUNTER(load_list)
                    return false;
                }
            }
        }
		sort (p.begin (), p.end ());
        ftime (&t2);
        fetchTime += 1000 * (t2.time - t1.time) + (t2.millitm - t1.millitm);
        if ((errno = dbcp->c_close (dbcp))) {
            cerr << strerror (errno) << "\n";
            STOP_COUNTER(load_list)
            return false;
        }
        // Joining lists
        *ptr = p;
        STOP_COUNTER(load_list)
        return true;
    } else {
        typedef  DataAndPos<off_t>   PPOS;
        typedef  vector<PPOS>        offList;

        offList                      tmpList[2];

        nc = 0;
        
        // Create an intersection of all lists (< MaxListCount)
        for (int i = 0; i <= min (CLen1Index::MaxIntersectCount, StrLen - NgramLen); i++) {
            char   mins[NgramLen + 2], maxs[NgramLen + 2];
            ngram  gmin, gmax;

            gmin.pack (&str[i], NgramLen, minpos + i);
            gmax.pack (&str[i], NgramLen, maxpos + i);

            gmin.toStr (mins, NgramLen);
            gmax.toStr (maxs, NgramLen);

            DBC *dbcp;
            DBT key, data;
            memset (&key, 0, sizeof (key));
            memset (&data, 0, sizeof (data));
            key.data = mins;
            key.size = NgramLen + 2;
            ftime (&t1);

            if ((errno = dbp->cursor (dbp, NULL, &dbcp, 0))) {
                cerr << strerror (errno) << "\n";
                STOP_COUNTER(load_list)
                return false;
            }

            unsigned int flags = DB_SET_RANGE;

            offList p;

            for (;;) {
                errno = dbcp->c_get (dbcp, &key, &data, flags);
                flags = DB_NEXT;
                if (errno) {
                    if (errno == DB_NOTFOUND)
                        break;
                    else {
                        cerr << strerror (errno) << "\n";
                        STOP_COUNTER(load_list)
                        return false;
                    }
                }
                // If greater than the last possible ngram than break
                if (strncmp ((char *) key.data, maxs, key.size) > 0)
                    break;
                // Now check whether pos is ok
                // Do not forget that we are incrementing acceptable pos value by variable i
                int pos = ngram::PosFromStr ((char *) key.data, NgramLen);
                if (pos <= maxpos + i && pos >= minpos + i) {
                    char tmp[NgramLen + 1];
                    strncpy(tmp, (char *) key.data, NgramLen);
                    tmp[NgramLen] = 0;

                    size_t listSize = *((off_t *) data.data);
                    size_t listOff = *(1 + (off_t *) data.data);
                    size_t listCnt = *(2 + (off_t *) data.data);

                    // Getting in-list offsets
                    char *ptr = reinterpret_cast<char *>(lstStart); 
                    ptr += listOff;

                    CMemBitReader      reader(ptr, listSize);

                    // Getting compressed in-list offsets
                    try {

		                p.reserve(p.size() + listCnt);

                        off_t last = 0;
                        off_t t;
                        if (!mUseDelta) {
                            for (int j = 0; j < listCnt; j++) {
                                VarByteDecode(t, reader);
                                t--;
                                last += t;
                                p.push_back(PPOS(last, pos - i));
                                recFetched++;
                            }
                        } else {
                            for (int j = 0; j < listCnt; j++) {
                                DeltaDecode(t, reader);
                                t--;
                                last += t;
                                p.push_back(PPOS(last, pos - i));
                                recFetched++;
                            }
                        }
                    } catch (const CReachedEof &) {
                        cerr << "unexpected EOF" << endl;
                        STOP_COUNTER(load_list)
                        return false;
                    }
                }
            }
			sort (p.begin (), p.end ());
            ftime (&t2);
            fetchTime +=
                1000 * (t2.time - t1.time) + (t2.millitm - t1.millitm);
            if ((errno = dbcp->c_close (dbcp))) {
                cerr << strerror (errno) << "\n";
                STOP_COUNTER(load_list)
                return false;
            }
            if (first) {
                first = false;
                tmpList[1 - nc] = p;
                nc = 1 - nc;
            } else {
                ListIntersection (tmpList[nc], p, &tmpList[1 - nc]);
                nc = 1 - nc;
            }
            if (tmpList[nc].size() <= 1) {
                break;
            }
        }
        //*ptr = pl[nc];
        ptr->reserve(tmpList[nc].size());
        for (unsigned int k = 0; k < tmpList[nc].size(); ++k)
            ptr->push_back(tmpList[nc][k].mData);
        STOP_COUNTER(load_list)
        return true;

    }
    cerr << "We mustn't reach this point (internal error)\n";
    STOP_COUNTER(load_list)
    return false;

}

void CLen1IndexBDB::ResetCounters()
{
    CLen1Index::ResetCounters();
    RESET_COUNTER(load_list);
}

bool CLen1IndexBDB::DumpDBCompress(const char *index_name, mOffset & offset_map)
{
    if (!opened) {
        cerr << index_name << " is not opened!" << endl;
        return false;
    }

    char        keystr[NgramLen + 2];
    DBT         key, data;
    size_t      startOff = 0;
    CBitWriter  bufferedWriter(1024 * 1024);

    memset (&key, 0, sizeof (key));
    memset (&data, 0, sizeof (data));
    key.data = (void *) keystr;
    key.size = NgramLen + 2;

    for (mPtrList::iterator it = invgram.begin (); it != invgram.end (); it++) {
        // Getting a key
        ngram g ((*it).first);
        g.toStr (keystr, NgramLen);
        PtrList *plist = invgram[g];
        int lsize = plist->size ();
        unsigned int keydat[3] = { 0, startOff, lsize };
        data.data = (void *) keydat;
        data.size = sizeof keydat;
        if (lsize > 0) {
            vector<off_t>    offsets(lsize);
            bufferedWriter.StartOutput(fd);
            PtrListIterator  it = plist->begin ();

            for (int i = 0; i < lsize; i++, it++) {
                offsets[i] = offset_map[*it];
            }
            
            try {
                sort(offsets.begin(), offsets.end());
#if 0
                cout <<  keyStr << " -> " << lsize << " avg: " << (offsets[lsize - 1] / double(lsize)) << endl;
#endif
                off_t last = 0;
                for (int i = 0; i < lsize; ++i) {
                    if (!mUseDelta) {
                        VarByteEncode(1 + offsets[i] - last, bufferedWriter);
                    } else {
                        DeltaEncode(1 + offsets[i] - last, bufferedWriter);
                    }
                    last = offsets[i];
                }
                bufferedWriter.FinishOutput();
            } catch (const std::runtime_error e) {
                cerr << "Error while writing lists: " << e.what() << endl;
                return false;
            }
        }

        off_t newStartOff = lseek(fd, 0, SEEK_CUR);

        if (newStartOff < 0) {
            Perror("Cannot lseek");
            return false;
        }
        keydat[0] = newStartOff - startOff + 1;
        startOff = newStartOff;
       
        errno = dbp->put (dbp, NULL, &key, &data, DB_NOOVERWRITE);
        if (errno) {
            cerr << (strerror (errno));
            return false;
        }
    }

    return true;
}

bool CLen1IndexBDB::DumpDBNoCompress(const char *index_name, mOffset & offset_map)
{
    if (!opened) {
        cerr << index_name << " is not opened!" << endl;
        return false;
    }
#if 0
    for (mPtrList::iterator it = invgram.begin ();
         it != invgram.end (); it++) {
        // Getting a key
        ngram g ((*it).first);
        PtrList *plist = invgram[g];
    }
#endif

    char   keystr[NgramLen + 2];
    DBT    key, data;
    size_t startOff = 0;

    memset (&key, 0, sizeof (key));
    memset (&data, 0, sizeof (data));
    key.data = (void *) keystr;
    key.size = NgramLen + 2;
    for (mPtrList::iterator it = invgram.begin ();
         it != invgram.end (); it++) {
        // Getting a key
        ngram g ((*it).first);
        g.toStr (keystr, NgramLen);
        PtrList *plist = invgram[g];
        int lsize = plist->size ();
        unsigned int keydat[2] = { lsize, startOff };
        data.data = (void *) keydat;
        data.size = sizeof keydat;
        if (lsize > 0) {
            vector<off_t>    offsets(lsize);
            PtrListIterator  it = plist->begin ();

            for (int i = 0; i < lsize; i++, it++) {
                offsets[i] = offset_map[*it];
                //cout<<" ## "<<keystr<<" #"<<(*plist)[i]<<offsets[i]<<"\n";
            }
            size_t wrSize = offsets.size() * sizeof offsets[0];

            if (write (fd, &offsets[0], wrSize) != wrSize) {
                Perror ("Cannot write to file");
                return false;
            }
            startOff += lsize;
        }
        errno = dbp->put (dbp, NULL, &key, &data, DB_NOOVERWRITE);
        if (errno) {
            cerr << (strerror (errno));
            return false;
        }
    }

    return true;
}

bool CLen1IndexBDB::ReadListDB(const char *str, int StrLen, int minpos,
                             int maxpos, OffsetList * ptr)
{
    return (mHasCompress) ? ReadListDBCompress(str, StrLen, minpos, maxpos, ptr):
                            ReadListDBNoCompress(str, StrLen, minpos, maxpos, ptr);
}

bool CLen1IndexBDB::DumpDB (const char *index_name, mOffset & offset_map)
{
    if (!opened) {
        cerr << index_name << " is not opened!" << endl;
        return false;
    }
    return (mHasCompress) ? DumpDBCompress(index_name, offset_map):DumpDBNoCompress(index_name, offset_map);
}

void CLen1IndexBDB::PrintAddStat(int testNum, int kerr, ostream & ostr)
{
    if (testNum <= 0) return;
    ostr << "kerr = " << kerr << " " <<
            "Time to fetch lists (ms): " << GET_COUNTER_MS(load_list) /testNum  <<  " "
            "Time to check strings (ms): " << GET_COUNTER_MS(check_string)  /testNum << "\n";
}

