#include <sys/time.h>
#include <iostream>

#include "interface.h"

#include "levendfa.h"


void usage(const char *prog)
{
    cerr << "Usage: " << prog << " 'string to create an neighborhood' 'kerr' 'string to compare'" << endl;
}

void GenNeighborhood(int start, string & str, bool hasTransp, int kerr, StringList * resList, const char *ABC)
{
    if (kerr <= 0) {
        if (resList) resList->push_back(str);
        return;
    }

    int size = str.size ();
    if (start > size) return;

    int i = 0;
    // Substitution
    for (i = start; i < size; i++) {
        char origChar = str[i];
        for (const char *ptr = ABC; *ptr; ptr++) {
            if (*ptr != str[i]) {
                str[i] = *ptr;
                GenNeighborhood(i + 1, str, hasTransp, kerr - 1, resList, ABC);
            }
        }
        str[i] = origChar;
    }
    // We well try to make as few insertions and erases as possible
    // It is enough to make 1 insertion, 1 deletion per insertion case
    // and 1 insertion, 1 deletion per deletion case
    // Insertion
    str.insert(str.begin() + start, ' ');
    for (i = start; i <= size; i++) {
        for (const char *ptr = ABC; *ptr; ptr++) {
            str[i] = *ptr; 
            GenNeighborhood(i + 1, str, hasTransp, kerr - 1, resList, ABC);
        }
        if (i < size) {
            swap(str[i], str[i + 1]);
        }
    }
    str.erase(str.end() - 1);
    if (hasTransp) {
    // Transposition
        for (i = start; i < size - 1; i++) {
            swap(str[i], str[i+1]);
            GenNeighborhood(i + 2, str, hasTransp, kerr - 1, resList, ABC);
            swap(str[i], str[i+1]);
        }
    }
    // Deletion
    char origChar = str[start];
    str.erase(str.begin() + start);
    for (i = start; i < size; i++) {
        GenNeighborhood(i, str, hasTransp, kerr - 1, resList, ABC);
        if (i < size - 1) {
           swap(str[i], origChar);
        }
    }
    str.insert (str.end(), origChar);
}


int main(int argc, char *argv[])
{
    if (argc != 4) {
        usage(argv[0]);
        return -1;
    }
    string     str(argv[1]);
    int        kerr = atol(argv[2]);
    timeval    t1, t2;
    StringList lst;

    gettimeofday(&t1, NULL);
    GenNeighborhood(0, str, true, kerr, &lst, EnglishABC);
    gettimeofday(&t2, NULL);
    cout << "Time (fill string list): " << ((t2.tv_sec - t1.tv_sec) * 1000000 + t2.tv_usec - t1.tv_usec) << endl;

    gettimeofday(&t1, NULL);
    GenNeighborhood(0, str, true, kerr, NULL, EnglishABC);
    gettimeofday(&t2, NULL);
    cout << "Time (no list to fill): " << ((t2.tv_sec - t1.tv_sec) * 1000000 + t2.tv_usec - t1.tv_usec) << endl;
    // Now gen leven dfa

    gettimeofday(&t1, NULL);
    LevenDFA dfa(EnglishABC, str, kerr, true);
    gettimeofday(&t2, NULL);
    cout << "Time to create DFA: " << ((t2.tv_sec - t1.tv_sec) * 1000000 + t2.tv_usec - t1.tv_usec) << endl;

    // Now some tests for searching
    string str1(argv[3]);
    int    sz1 = str1.size();
    gettimeofday(&t1, NULL);
    for(int i =0 ; i < 1000; ++i) {
        dfa.Start();
        bool final;
        for (int k = 0; k < sz1; ++k) {
           if (!dfa.Next(str1[k], final)) break;
        }
    }
    gettimeofday(&t2, NULL);
    cout << "Time to search in DFA: " << (((t2.tv_sec - t1.tv_sec) * 1000000 + t2.tv_usec - t1.tv_usec)) / 1000.0 << " microseconds " << endl;
}
