#include "levendfa.h"
#include <stdexcept>
#include <iostream>
#include "../Lang/langcase.h"

static int memSize = 0;

void LevenDFA::allocateNodesFull(int N)
{ 
    unsigned int sz = N * m_nodeSizeFull;
    char *p = (char *)malloc(sz);
    memset(p, 0, sz);
    if (!p) {
        throw bad_alloc();
    }
    m_memBlocks.push_back(p);
    for (int i = 0; i < N; ++i) {
        m_freeNodesFull.insert(m_freeNodesFull.end(), (LevenNodeFull *)p);
        p += m_nodeSizeFull;
    }
    memSize += N * m_nodeSizeFull;
    //cout << "Number of bytes allocated: " << memSize << endl;
}

void LevenDFA::allocateNodesShort(int N)
{ 
    unsigned int sz = N * m_nodeSizeShort;
    char *p = (char *)malloc(sz);
    memset(p, 0, sz);
    if (!p) {
        throw bad_alloc();
    }
    m_memBlocks.push_back(p);
    for (int i = 0; i < N; ++i) {
        m_freeNodesShort.insert(m_freeNodesShort.end(), (LevenNodeShort *)p);
        p += m_nodeSizeShort;
    }
    memSize += N * m_nodeSizeShort;
    //cout << "Number of bytes allocated: " << memSize << endl;
}

LevenDFA::LevenDFA(const char *pABC, const string & str, int kerr, bool hasTransp, bool exDist) : m_kerr(kerr), m_hasTransp(hasTransp), m_exDist(exDist)
{
    m_abcSize = 0; 
    // Prepare the map first
    memset(m_abcMap, 0, sizeof m_abcMap);
    strncpy(m_ABC, pABC, sizeof m_ABC);
    for (const char *p = pABC; *p; ++p) {
        unsigned char c1 = tolower_eng_rus(*p);
        unsigned char c2 = toupper_eng_rus(*p);
        m_abcMap[c1] = m_abcMap[c2] = ++m_abcSize;
    }
    m_abcMap[EOS] = ++m_abcSize;
    m_nodeSizeFull = sizeof(LevenNodeFull) + sizeof(LevenNodeBase *) * (m_abcSize + 1);
    m_nodeSizeShort = sizeof(LevenNodeShort);

    string s = str;
    s += EOS;
    for (int i = 0; i < str.size(); ++i) {
        s[i] = toupper_eng_rus(s[i]);
    }

    m_pCurr = m_pRoot = getNewNodeShort(0);
    GenNeighborhood(&m_pRoot, 0, s, kerr, hasTransp);
}

LevenDFA::LevenNodeFull *LevenDFA::getNewNodeFull(unsigned char c)
{
    list<LevenNodeFull *>::iterator it = m_freeNodesFull.begin();
    if (it == m_freeNodesFull.end()) {
        allocateNodesFull(100);
        it = m_freeNodesFull.begin();
        assert(it != m_freeNodesFull.end());
    }

    LevenNodeFull *p = *it;
    p->mIsFull = true;
    p->mSym = c;
    m_freeNodesFull.erase(it);
    return p;
}

LevenDFA::LevenNodeShort *LevenDFA::getNewNodeShort(unsigned char c)
{
    list<LevenNodeShort *>::iterator it = m_freeNodesShort.begin();
    if (it == m_freeNodesShort.end()) {
        allocateNodesShort(100);
        it = m_freeNodesShort.begin();
        assert(it != m_freeNodesShort.end());
    }

    LevenNodeShort *p = *it;
    p->mSym = c;
    m_freeNodesShort.erase(it);
    return p;
}

LevenDFA::~LevenDFA()
{
    list<void *>::iterator endIt = m_memBlocks.end();
    for (list<void *>::iterator it = m_memBlocks.begin();it != endIt; ++it) {
        free(*it);
    }
}

void LevenDFA::GenNeighborhood(LevenNodeBase **ppStartNode, int start, string & str, int kerr, bool hasTransp)
{
    if (kerr <= 0) {
        InsertString(ppStartNode, str, start, kerr);
        return;
    }

    int size = str.size ();
    if (start > size) return;

    int i = 0;
    // Last character EOS is never deleted/replaced or transposed!!!
    assert(size >= 1);
    // Substitution
     
    if (!GetParentPtr(*ppStartNode, str[start])) {
        InsertString(ppStartNode, str, start, kerr);
    }
    // Replacement
    LevenNodeBase **ppCurr = ppStartNode;
    for (i = start; i < size - 1; i++) {
        assert(ppCurr);
        char origChar = str[i];
        for (const char *ptr = m_ABC; *ptr; ptr++) {
            if (*ptr != str[i]) {
                str[i] = *ptr;
                InsertString(ppCurr, str, i, kerr - 1);
                LevenNodeBase **ppCurr1 = GetParentPtr(*ppCurr, str[i]);
                assert(ppCurr1);
                if (kerr > 1) {
                    GenNeighborhood(ppCurr1, i + 1, str, kerr - 1, hasTransp);
                }
            }
        }
        str[i] = origChar;
        ppCurr = GetParentPtr(*ppCurr, str[i]);
        assert(ppCurr);
    }
    // We well try to make as few insertions and deletions as possible
    // It is enough to make 1 insertion, 1 deletion per insertion case
    // and 1 insertion, 1 deletion per deletion case
    // Insertion
    str.insert(str.begin() + start, ' ');
    ppCurr = ppStartNode;
    for (i = start; i < size; i++) {
        assert(ppCurr);
        for (const char *ptr = m_ABC; *ptr; ptr++) {
            str[i] = *ptr; 
            InsertString(ppCurr, str, i, kerr - 1);
            LevenNodeBase **ppCurr1 = GetParentPtr(*ppCurr, str[i]);
            assert(ppCurr1);
            if (kerr > 1) {
                GenNeighborhood(ppCurr1, i + 1, str, kerr - 1, hasTransp);
            }
        }
        swap(str[i], str[i + 1]);
        
        ppCurr = GetParentPtr(*ppCurr, str[i]);
        assert(ppCurr);
    }
    str.erase(str.end() - 1);
    if (hasTransp) {
    // Transpositions
        ppCurr = ppStartNode;
        for (i = start; i < size - 2; i++) {
            swap(str[i], str[i+1]);
            InsertString(ppCurr, str, i, kerr - 1);
            LevenNodeBase **ppCurr1 = GetParentPtr(*ppCurr, str[i]);
            assert(ppCurr1);
            LevenNodeBase **ppCurr2 = GetParentPtr(*ppCurr1, str[i + 1]);
            assert(ppCurr2);
            if (kerr > 1) {
                GenNeighborhood(ppCurr2, i + 2, str, kerr - 1, hasTransp);
            }
            swap(str[i], str[i+1]);
            ppCurr = GetParentPtr(*ppCurr, str[i]);
            assert(ppCurr);
        }
    }
 
    // Deletions
    if (start < size - 1) {
        char origChar = str[start];
#ifndef NDEBUG
        string s1 = str;
#endif
        str.erase(str.begin() + start);
        ppCurr = ppStartNode;
        for (i = start; i < size - 1; i++) {
            assert(ppCurr);
            InsertString(ppCurr, str, i, kerr - 1);
            if (kerr > 1) {
                GenNeighborhood(ppCurr, i, str, kerr - 1, hasTransp);
            }
            swap(str[i], origChar);
            ppCurr = GetParentPtr(*ppCurr, str[i]);
            assert(ppCurr);
        }
        str.insert(str.begin() + size - 1, origChar);
        assert(str == s1);
    }
}
