#include <assert.h>
#include <iostream>
#include <algorithm>
#include "fqnodemultipivot.h"

#ifdef MYERS_DIST
    #include "myers.hpp"
#else
    #include "edist.h"
#endif
#include "myagrep.h"


#ifdef MYERS_DIST
CMyersEdistFastPair EdCalc;
#endif


// Do not change this
const char CFQNodeMultiPivot::termChar = 0;

void CFQNodeMultiPivot::Insert(const StringList & lst, bool hasTransp, bool silent, int level)
{
    if (!silent) {
        std:: cout << " Insert of " << lst.size () << " strings, level = "<<level<<" this = "<<this << std::endl;
    }
    // First of all we choose the root randomly if not exists
    unsigned int ListSize = lst.size ();
    if (!ListSize) {
        return;
    }
    // If the number of strings is less than or equal to the maximal number of pivots,
    // just store strings as pivots
    if (lst.size () <= pivotCnt) {
        pivots = lst;
        return;
    }
    // Choose random & unique pivots
    if (!pivots.size()) {
        
        for (int k = 0; k < pivotCnt; ++k) {
            int num = 0;
        
            string p;
            do {
                num = rand () % ListSize;

                p = lst[num];
            } while (find(pivots.begin(), pivots.end(), p) != pivots.end());

            pivots.push_back(p); 
        }
    }

    vector<pair<vus, string> > SplitList;

    for (StringList::const_iterator it = lst.begin (); it != lst.end (); ++it) {

        vus dists(pivotCnt);

        bool HasEqual = false;
        for (int k = 0; k < pivotCnt; ++k) {
#ifdef MYERS_DIST
            if (!(dists[k] = (u_int16_t) EdCalc.CalcDist(pivots[k].c_str(), it->c_str(), hasTransp))) {
#else
            if (!(dists[k] = (u_int16_t) EditDistance(pivots[k], *it, hasTransp))) {
#endif
                HasEqual = true;
                break;
            }
        }

        if (!HasEqual) {
            if (dists[0] >= nodes.size()) {
                nodes.resize(dists[0] + 1);
            }
            SplitList.push_back(make_pair(dists, *it));
        }
    }
    if (!SplitList.size()) return;

    sort(SplitList.begin(), SplitList.end()); // Sort the list by the distances from pivots


    StringList tmpList;
    vus  prev = SplitList[0].first;

    for (int i = 0; i < SplitList.size(); ++i) {
        if (prev != SplitList[i].first) {
            AddTmpList(tmpList, prev, hasTransp, silent, level);
            prev = SplitList[i].first;
        }
        tmpList.push_back(SplitList[i].second);
    }
    if (tmpList.size()) {
        AddTmpList(tmpList, prev, hasTransp, silent, level);
    }
}

void CFQNodeMultiPivot::AddTmpList(StringList &tmpList, vus &prev, bool hasTransp, bool silent, int level)
{
    CFQNodeMultiPivot   *pChild = new CFQNodeMultiPivot(pivotCnt);
    pChild->Insert(tmpList, hasTransp, silent, level + 1);
    int dst1 = prev[0];
    prev.erase(prev.begin());
    nodes[dst1].push_back(OneChild(pChild, prev));
    tmpList.clear();
}

int CFQNodeMultiPivot::DumpToFile(off_t offset, int fd)
{
    return 1;
}

void CFQNodeMultiPivot::SearchAtLevel (const std::string & s, int nerr, const myagrep & a,
                             int *numComp, StringList * resList, bool hasTransp,
                             bool silent)
{
    if (numComp)
        (*numComp)++;
    if (!nodes.size() && bucket.size()) {
        if (numComp)
            (*numComp) += bulkCnt;
        a.findPattern(CInterface::filterType, termChar,
                          bucket.data(), bucket.data() + bucket.size(), silent, resList);
        return;
    }

    // We must check the root
    u_int16_t dists[pivots.size()];

    int last = 0;

    for (int i = 0; i < pivots.size(); ++i) {
#ifdef MYERS_DIST
        dists[i] = EdCalc.CalcDist(s.c_str(), pivots[i].c_str(), hasTransp);
#else
        dists[i] = EditDistance (s, pivots[i], hasTransp);
#endif
        if (dists[i] <= nerr) {
            if (!silent)
                cout << pivots[i] << "\n";
            if (resList != NULL)
                resList->push_back(pivots[i]);
        }
    }
    if (!nodes.size()) {
        return; // No children
    }
    int imin = dists[0] - nerr, imax = dists[0] + nerr;
    if (imin < 1)
        imin = 1;
    if (imax > nodes.size() - 1)
        imax = nodes.size() - 1;

    for (int i = imin; i <= imax; i++) {
        Children &chlds = nodes[i];
        int      N = chlds.size();
        for (int k = 0; k < N; ++k) {
            OneChild &achild = chlds[k];
            bool hasMatch = true;
            for (int m = 1; m < pivots.size(); ++m) {
                if (abs(dists[m] - achild.dists[m - 1]) > nerr) {
                    hasMatch = false;
                    break;
                }
            }
            if (hasMatch) {
                achild.pChild->SearchAtLevel (s, nerr, a, numComp, resList, hasTransp, silent);
            }
        }
    }

}

void CFQNodeMultiPivot::Print (int level)
{
    for (int i = 0; i < level; i++) {
        cout << "*";
    }
    cout << ":" << bucket;
    for (int k = 0; k < pivots.size(); ++k) {
        cout << pivots[k]  << ";";
    }
    cout << "\n";
    for (int i = 1; i < nodes.size(); i++) {
        for (int k = 0; k < nodes[i].size(); ++k) {
            for (int j = 0; j < level; j++) {
                cout << "  ";
            }
            cout << " dst (" << i;
            for (int m = 0; m < nodes[i][k].dists.size(); ++m)
                cout << "," << nodes[i][k].dists[m];
            cout << ")" << endl;
            nodes[i][k].pChild->Print (level + 1);
        }
    }
}

void CFQNodeMultiPivot::appendAllStrings(string & str)
{
    for (int k = 0; k < pivots.size(); ++k) {
        str.append(pivots[k].data(), pivots[k].size());
        str.append(&termChar, 1);
    }
    //cout << "Appedning of size " << root.size() << endl;
    for (int i = 1; i < nodes.size(); ++i) 
    for (int k = 0; k < nodes[i].size(); ++k) {
        nodes[i][k].pChild->appendAllStrings(str);
    }
}

void CFQNodeMultiPivot::CreateBuckets(int buckSize, int level)
{
    if (!level) fillTreeCnt();

    buckSize = max(buckSize, 1);

    unsigned int cnt = 0;
    for (int i = 1; i < nodes.size(); ++i) {
        cnt += nodes[i].size();
    }

    if (cnt && treeCnt > 1 && // this is important, only then bulkCnt will be greater than 1
         treeCnt <= buckSize) {
        appendAllStrings(bucket); // Concatenate all strings into one
        // Free all children
        for (int i = 1; i < nodes.size(); ++i) 
        for (int k = 0; k < nodes[i].size(); ++k) {
            delete nodes[i][k].pChild;
        }

        nodes.resize(0);
        pivots.resize(0);

        bulkCnt = count(bucket.begin(), bucket.end(), termChar);
    } else {
        for (int i = 1; i < nodes.size(); ++i) 
        for (int k = 0; k < nodes[i].size(); ++k) {
            nodes[i][k].pChild->CreateBuckets(buckSize, level + 1);
        }
    }
} 

void CFQNodeMultiPivot::fillTreeCnt()
{
    treeCnt = pivots.size();

    for (int i = 1; i < nodes.size(); ++i) 
    for (int k = 0; k < nodes[i].size(); ++k) {
        nodes[i][k].pChild->fillTreeCnt();
        treeCnt += nodes[i][k].pChild->treeCnt;
    }
}

CFQNodeMultiPivot::~CFQNodeMultiPivot () {
    for (int i = 1; i < nodes.size(); i++)
        for (int k = 0; k < nodes[i].size(); ++k) {
            delete nodes[i][k].pChild;
        }
}
