#include <algorithm>
#include "index.h"

using namespace std;

CTrieIndex::CTrieIndex (bool hasTransp, EAlphabet AlphaType)
            : CInterface(hasTransp, AlphaType), opened (false), fdFwd(0), fdRev(0)
{
}

CTrieIndex::~CTrieIndex ()
{
    if (opened)
        CloseDB ();
}

bool CTrieIndex::Insert (const char *str)
{
    try {
        string s(str);

        nodeFwd.Insert(s + EOS);
        reverse(s.begin(), s.end());

        nodeRev.Insert(s + EOS);

    } catch (const string & e) {
        cerr << "Error inserting of '" << (string(str) +EOS) << "': " << e << endl;
        throw;
    }
    return true;
}

bool CTrieIndex::Search(const char *str, int kerr, int *numComp,
                         StringList * resList, bool silent)
{
    string s(str);


    if (kerr != 0) {
        int switchLen = s.size() / 2;
        int switchLen1 = s.size() - switchLen;

        if (switchLen <= 0) {
            nodeFwd.Search(s, kerr, numComp, resList, mHasTransp, silent);
            return true;
        }

        StringList     tmpList, tmpListReversed;

            // Process a specific transposition case
            if (mHasTransp) {
                swap(s[switchLen - 1], s[switchLen]); // switchLen > 0

                if (kerr == 1) {
                    nodeFwd.SearchExact(s, numComp, &tmpList, silent);
                } else {
                    int kerr1 = 0, kerr2 = kerr - 1;
                    for (; kerr2 >= kerr1; ++kerr1, --kerr2) {
                        nodeFwd.SearchSplit(s, kerr1, true, kerr2, false, numComp, switchLen, &tmpList, mHasTransp, silent);
                    }

                    reverse(s.begin(), s.end());
                    for (; kerr2 >= 0; ++kerr1, --kerr2) {
                        nodeRev.SearchSplit(s, kerr2, false, kerr1, true, numComp, switchLen1, &tmpListReversed, mHasTransp, silent);
                    }
                    reverse(s.begin(), s.end());
                }


                swap(s[switchLen - 1], s[switchLen]); // switchLen > 0
            }
 
            int kerr1 = 0, kerr2 = kerr;
            for (; kerr2 >= kerr1; ++kerr1, --kerr2) {
                nodeFwd.SearchSplit(s, kerr1, true, kerr2, false, numComp, switchLen, &tmpList, mHasTransp, silent);
            }

            reverse(s.begin(), s.end());
            for (; kerr2 >= 0; ++kerr1, --kerr2) {
                nodeRev.SearchSplit(s, kerr2, false, kerr1, true, numComp, switchLen1, &tmpListReversed, mHasTransp, silent);
            }
            reverse(s.begin(), s.end());
                       

            for (int i = 0; i < tmpListReversed.size(); ++i) {
               string &s = tmpListReversed[i];
               reverse(s.begin(), s.end());
               tmpList.push_back(s);
            }

            sort(tmpList.begin(), tmpList.end());
            tmpList.erase(unique(tmpList.begin(), tmpList.end()), tmpList.end());
            if (resList) {
                *resList = tmpList;
            }
    } else {
        nodeFwd.SearchExact(s, numComp, resList, silent);
    }
    return true;
}

bool CTrieIndex::SearchInDB (const char *str, int kerr, int *numComp,
                             StringList * resList, 
                             bool silent)
{
    if (!opened) {
        cerr << "Database is not opened!" << endl;
    }

    std::string s(str);

    if (kerr != 0) {
        int switchLen = s.size() / 2;
        int switchLen1 = s.size() - switchLen;

        if (switchLen <= 0) {
            if (!CTrie::SearchInFile(fdFwd, s, kerr, numComp, resList, mHasTransp, silent)) return false;
            return true;
        }

        StringList     tmpList, tmpListReversed;

        // Process a specific transposition case
        if (mHasTransp) {
            swap(s[switchLen - 1], s[switchLen]); // switchLen > 0

            if (kerr == 1) {
                if (!CTrie::SearchInFileExact(fdFwd, s, numComp, &tmpList, silent)) return false;
            } else {
                int kerr1 = 0, kerr2 = kerr - 1;
                for (; kerr2 >= kerr1; ++kerr1, --kerr2) {
                    if (!CTrie::SearchInFileSplit(fdFwd, s, kerr1, true, kerr2, false, 
                                numComp, switchLen, &tmpList, mHasTransp, silent)) return false;
                }
                reverse(s.begin(), s.end());
                for (; kerr2 >= 0; ++kerr1, --kerr2) {
                    if (!CTrie::SearchInFileSplit(fdRev, s, kerr2, false, kerr1, true, 
                                numComp, switchLen1, &tmpListReversed, mHasTransp, silent)) return false;
                }
                reverse(s.begin(), s.end());
            }

            swap(s[switchLen - 1], s[switchLen]); // switchLen > 0
        }

        int kerr1 = 0, kerr2 = kerr;
        for (; kerr2 >= kerr1; ++kerr1, --kerr2) {
            if (!CTrie::SearchInFileSplit(fdFwd, s, kerr1, true, kerr2, false, 
                        numComp, switchLen, &tmpList, mHasTransp, silent)) return false;
        }
        reverse(s.begin(), s.end());
        for (; kerr2 >= 0; ++kerr1, --kerr2) {
            if (!CTrie::SearchInFileSplit(fdRev, s, kerr2, false, kerr1, true, 
                        numComp, switchLen1, &tmpListReversed, mHasTransp, silent)) return false;
        }
        reverse(s.begin(), s.end());


        for (int i = 0; i < tmpListReversed.size(); ++i) {
           string &s = tmpListReversed[i];
           reverse(s.begin(), s.end());
           tmpList.push_back(s);
        }

        sort(tmpList.begin(), tmpList.end());
        tmpList.erase(unique(tmpList.begin(), tmpList.end()), tmpList.end());

        if (resList) {
            *resList = tmpList;
        }
    } else {
            if (!CTrie::SearchInFileExact(fdFwd, s, numComp, resList, silent)) return false;
    }
    return true;
}

bool CTrieIndex::OpenDB (const char *file_name, bool create)
{
    string fwdName = string(file_name) + "_fwd.img";
    string revName = string(file_name) + "_rev.img";

    if (opened)
        CloseDB ();
    if (!create) {
        fdFwd = open (fwdName.c_str(), O_RDONLY, 0);
        if (fdFwd <= 0) {
            Perror ("Error opening file");
            return false;
        }
        fdRev = open (revName.c_str(), O_RDONLY, 0);
        if (fdRev <= 0) {
            Perror ("Error opening file");
            return false;
        }
    } else {
        fdFwd = open (fwdName.c_str(), O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
        if (fdFwd <= 0) {
            Perror ("Error opening file");
            return false;
        }
        fdRev = open (revName.c_str(), O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
        if (fdRev <= 0) {
            Perror ("Error opening file");
            return false;
        }
    }
    return (opened = true);
}

bool CTrieIndex::CloseDB ()
{
    if (!opened)
        return true;
    if (close(fdFwd) || close(fdRev)) {
        return false;
    }
    opened = false;
    return true;
}


bool CTrieIndex::DumpDB (const char *file_name)
{
    bool stat;
    if (opened)
        CloseDB ();
    if (!OpenDB (file_name, true)) {
        return false;
    }
    //node.Print();
    stat = nodeFwd.DumpToFile(fdFwd, 0) && nodeRev.DumpToFile(fdRev, 0);
    return stat && CloseDB ();
}

#ifdef _DEBUG
void CTrieIndex::Print ()
{
    node.Print ();
}
#endif
